import importlib.util
import os
import tempfile
import time
from concurrent.futures import ThreadPoolExecutor

import numpy as np
import pytest
import xarray as xr

from xyzpy.gen.cropping import grow
from xyzpy.gen.farming import Harvester, Runner, Sampler, label
from xyzpy.manage import load_df, load_ds

found_h5netcdf = importlib.util.find_spec("h5netcdf") is not None
found_netcdf4 = importlib.util.find_spec("netcdf4") is not None
found_zarr = importlib.util.find_spec("zarr") is not None

h5netcdf_case = pytest.param(
    "h5netcdf",
    marks=pytest.mark.skipif(
        not found_h5netcdf, reason="h5netcdf not installed"
    ),
)
netcdf4_case = pytest.param(
    "netcdf4",
    marks=pytest.mark.skipif(
        not found_netcdf4, reason="netcdf4 not installed"
    ),
)
zarr_case = pytest.param(
    "zarr",
    marks=pytest.mark.skipif(not found_zarr, reason="zarr not installed"),
)


# -------------------------------- Fixtures --------------------------------- #


def fn1_x(x):
    return x**2


def fn3_fba(a, b, c):
    sm = a + b + c
    ev = (a + b + c) % 2 == 0
    ts = a * (b * np.linspace(0, 1.0, 3) + c)
    return sm, int(ev), ts


@pytest.fixture
def fn3_fba_runner():
    r = Runner(
        fn3_fba,
        fn_args=("a", "b"),
        var_names=["sum", "even", "array"],
        var_dims={"array": ["time"]},
        var_coords={"time": np.linspace(0, 1.0, 3)},
        constants={"c": 100},
        attrs={"fruit": "apples"},
    )
    return r


@pytest.fixture
def fn3_fba_ds():
    return xr.Dataset(
        coords={"a": [1, 2], "b": [3, 4], "time": np.linspace(0, 1.0, 3)},
        data_vars={
            "sum": (("a", "b"), [[104, 105], [105, 106]]),
            "even": (("a", "b"), [[True, False], [False, True]]),
            "array": (
                ("a", "b", "time"),
                [
                    [[100, 101.5, 103], [100, 102, 104]],
                    [[200, 203, 206], [200, 204, 208]],
                ],
            ),
        },
        attrs={"c": 100, "fruit": "apples"},
    )


# ------------------------------ Runner Tests ------------------------------- #


class TestRunner:
    def test_basic(self):
        r = Runner(fn1_x, "x^2")
        print(str(r))
        print(repr(r))
        assert r(2) == 4

    @pytest.mark.parametrize(
        "prop, vals",
        [
            ("var_names", ["S", "E", "A"]),
            ("fn_args", ["A", "B", "C"]),
            ("var_dims", {"array": ["t"]}),
            ("var_coords", {"time": [1, 2]}),
            ("constants", [("c", 200)]),
            ("resources", [("c", 300)]),
        ],
    )
    def test_properties(self, prop, vals):
        r = Runner(
            fn3_fba,
            fn_args=("a", "b", "c"),
            var_names=["sum", "even", "array"],
            var_dims={"array": ["time"]},
            var_coords={"time": np.linspace(0, 1.0, 3)},
            constants={"c": 100},
            attrs={"fruit": "apples"},
        )
        getattr(r, prop)
        setattr(r, prop, vals)
        delattr(r, prop)
        assert getattr(r, prop) is None

    @pytest.mark.parametrize("var_names", ["sq", ("sq",), ["sq"]])
    @pytest.mark.parametrize("var_dims", [[], (), {}, {"sq": ()}])
    @pytest.mark.parametrize(
        "combos",
        [
            ("x", (1, 2, 3)),
            (("x", (1, 2, 3)),),
            ("x", [1, 2, 3]),
            [("x", [1, 2, 3])],
            {"x": (1, 2, 3)},
            {"x": [1, 2, 3]},
        ],
    )
    def test_runner_combos_parse(self, var_names, var_dims, combos):
        r = Runner(fn1_x, var_names=var_names, var_dims=var_dims)
        r.run_combos(combos)
        expected_ds = xr.Dataset(
            coords={"x": [1, 2, 3]}, data_vars={"sq": ("x", [1, 4, 9])}
        )
        assert r.last_ds.identical(expected_ds)

    def test_runner_combos(self, fn3_fba_runner, fn3_fba_ds):
        r = fn3_fba_runner
        r.run_combos((("a", (1, 2)), ("b", (3, 4))))
        assert r.last_ds.identical(fn3_fba_ds)

    def test_sow_reap_seperate(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            r = fn3_fba_runner
            crop = r.Crop(parent_dir=tmpdir, num_batches=2)
            crop.sow_combos((("a", (1, 2)), ("b", (3, 4))))
            for i in [1, 2]:
                grow(i, crop)
            crop.reap()
            assert r.last_ds.identical(fn3_fba_ds)

    def test_sow_and_reap(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            r = fn3_fba_runner
            crop = r.Crop(parent_dir=tmpdir, num_batches=2)

            def concurrent_grow():
                # wait for cases to be sown
                time.sleep(0.5)
                for i in [1, 2]:
                    grow(i, crop)

            with ThreadPoolExecutor(1) as pool:
                pool.submit(concurrent_grow)

                crop.sow_combos((("a", (1, 2)), ("b", (3, 4))))
                crop.reap(wait=True)

            assert r.last_ds.identical(fn3_fba_ds)

    @pytest.mark.parametrize("cases", [(1, 2, 3), ([1], [2], [3])])
    def test_runner_cases_parse(self, cases):
        r = Runner(fn1_x, var_names="sq", fn_args="x")
        r.run_cases(cases)
        expected_ds = xr.Dataset(
            coords={"x": [1, 2, 3]}, data_vars={"sq": ("x", [1, 4, 9])}
        )
        assert r.last_ds.identical(expected_ds)

    @pytest.mark.parametrize("dict_cases", [False, True])
    def test_runner_cases(self, fn3_fba_runner, dict_cases):
        if dict_cases:
            cases = [{"a": 2, "b": 3}, {"a": 1, "b": 4}]
        else:
            cases = [(2, 3), (1, 4)]

        fn3_fba_runner.run_cases(cases)
        expected_ds = xr.Dataset(
            coords={"a": [1, 2], "b": [3, 4], "time": np.linspace(0, 1.0, 3)},
            data_vars={
                "sum": (("a", "b"), [[np.nan, 105], [105, np.nan]]),
                "even": (("a", "b"), [[None, False], [False, None]]),
                "array": (
                    ("a", "b", "time"),
                    [
                        [[np.nan, np.nan, np.nan], [100, 102, 104]],
                        [[200, 203, 206], [np.nan, np.nan, np.nan]],
                    ],
                ),
            },
            attrs={"c": 100, "fruit": "apples"},
        )
        assert fn3_fba_runner.last_ds.identical(expected_ds)


# -------------------------- Harvester Tests -------------------------------- #


class TestHarvester:
    def test_save_and_load_ds(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, "test.h5")
            h = Harvester(fn3_fba_runner, fl_pth, full_ds=fn3_fba_ds)
            h.save_full_ds()
            h.load_full_ds()
            assert h.full_ds.equals(fn3_fba_ds)

    def test_harvest_combos_new(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, "test.h5")
            h = Harvester(fn3_fba_runner, fl_pth)
            h.harvest_combos((("a", (1, 2)), ("b", (3, 4))))
            hds = load_ds(fl_pth)
        assert h.last_ds.identical(fn3_fba_ds)
        assert h.full_ds.identical(fn3_fba_ds)
        assert hds.identical(fn3_fba_ds)

    def test_label_as_harvester(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, "test.h5")
            h = label(
                harvester=fl_pth,
                fn_args=("a", "b"),
                var_names=["sum", "even", "array"],
                var_dims={"array": ["time"]},
                var_coords={"time": np.linspace(0, 1.0, 3)},
                constants={"c": 100},
                attrs={"fruit": "apples"},
            )(fn3_fba)
            h.harvest_combos((("a", (1, 2)), ("b", (3, 4))))
            hds = load_ds(fl_pth)
        assert h.last_ds.identical(fn3_fba_ds)
        assert h.full_ds.identical(fn3_fba_ds)
        assert hds.identical(fn3_fba_ds)

    def test_harvest_combos_new_sow_reap_separate(
        self, fn3_fba_runner, fn3_fba_ds
    ):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, "test.h5")
            h = Harvester(fn3_fba_runner, fl_pth)
            crop = h.Crop(parent_dir=tmpdir, num_batches=2)

            crop.sow_combos((("a", (1, 2)), ("b", (3, 4))))

            for i in [1, 2]:
                grow(i, crop)
            crop.reap()

            hds = load_ds(fl_pth)

            assert h.last_ds.identical(fn3_fba_ds)
            assert h.full_ds.identical(fn3_fba_ds)
            assert hds.identical(fn3_fba_ds)

    def test_harvest_combos_sow_and_reap_thread_wait(
        self, fn3_fba_runner, fn3_fba_ds
    ):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, "test.h5")
            h = Harvester(fn3_fba_runner, fl_pth)
            crop = h.Crop(parent_dir=tmpdir, num_batches=2)

            def concurrent_grow():
                # wait for cases to be sown
                time.sleep(0.5)
                for i in [1, 2]:
                    grow(i, crop)

            with ThreadPoolExecutor(1) as pool:
                pool.submit(concurrent_grow)
                crop.sow_combos((("a", (1, 2)), ("b", (3, 4))))
                crop.reap(wait=True)

            hds = load_ds(fl_pth)

        assert h.last_ds.identical(fn3_fba_ds)
        assert h.full_ds.identical(fn3_fba_ds)
        assert hds.identical(fn3_fba_ds)

    def test_harvest_combos_new_sow_reap_incomplete(
        self, fn3_fba_runner, fn3_fba_ds
    ):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, "test.h5")
            h = Harvester(fn3_fba_runner, fl_pth)
            crop = h.Crop(parent_dir=tmpdir, num_batches=3)
            crop.sow_combos((("a", (1, 2)), ("b", (3, 4))))

            grow(1, crop)
            crop.reap(allow_incomplete=True)
            assert not h.full_ds.identical(fn3_fba_ds)
            hds = load_ds(fl_pth)
            assert not hds.identical(fn3_fba_ds)

            crop.grow_missing()
            crop.reap(allow_incomplete=True)
            h.load_full_ds()
            assert h.full_ds.equals(fn3_fba_ds)
            hds = load_ds(fl_pth)
            assert hds.identical(fn3_fba_ds)

    def test_harvest_combos_merge(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, "test.h5")
            h = Harvester(fn3_fba_runner, fl_pth)
            h.harvest_combos((("a", (1,)), ("b", (3, 4))))
            h.harvest_combos((("a", (2,)), ("b", (3, 4))))
            hds = load_ds(fl_pth)
        assert not h.last_ds.identical(fn3_fba_ds)
        assert h.full_ds.identical(fn3_fba_ds)
        assert hds.identical(fn3_fba_ds)

    def test_harvest_combos_overwrite(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, "test.h5")
            mod_ds = fn3_fba_ds.copy(deep=True)
            mod_ds["array"].loc[{"a": 1, "b": 3}] = 999
            h = Harvester(fn3_fba_runner, fl_pth, full_ds=mod_ds)
            h.save_full_ds()
            assert not h.full_ds.equals(fn3_fba_ds)
            h.harvest_combos((("a", (1,)), ("b", (3,))), overwrite=True)
            assert h.full_ds.equals(fn3_fba_ds)

    @pytest.mark.parametrize("dict_cases", [False, True])
    def test_harvest_cases_new(self, fn3_fba_runner, fn3_fba_ds, dict_cases):
        if dict_cases:
            cases = [
                {"a": 1, "b": 3},
                {"a": 1, "b": 4},
                {"a": 2, "b": 3},
                {"a": 2, "b": 4},
            ]
        else:
            cases = [(1, 3), (1, 4), (2, 3), (2, 4)]

        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, "test.h5")
            h = Harvester(fn3_fba_runner, fl_pth)
            h.harvest_cases(cases)
            hds = load_ds(fl_pth)
        assert h.last_ds.identical(fn3_fba_ds)
        assert h.full_ds.identical(fn3_fba_ds)
        assert hds.identical(fn3_fba_ds)

    def test_harvest_cases_merge(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, "test.h5")
            h = Harvester(fn3_fba_runner, fl_pth)
            h.harvest_cases([(1, 3), (2, 4)])
            h.harvest_cases([(1, 4), (2, 3)])
            hds = load_ds(fl_pth)
        assert not h.last_ds.identical(fn3_fba_ds)
        assert h.full_ds.identical(fn3_fba_ds)
        assert hds.identical(fn3_fba_ds)

    def test_harvest_cases_overwrite(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, "test.h5")
            mod_ds = fn3_fba_ds.copy(deep=True)
            mod_ds["array"].loc[{"a": 1, "b": 3}] = 999
            h = Harvester(fn3_fba_runner, fl_pth, full_ds=mod_ds)
            h.save_full_ds()
            assert not h.full_ds.equals(fn3_fba_ds)
            h.harvest_cases([(1, 3)], overwrite=True)
            assert h.full_ds.equals(fn3_fba_ds)

    @pytest.mark.parametrize(
        "engine", [h5netcdf_case, netcdf4_case, zarr_case]
    )
    def test_harvest_cases_merge_dask(
        self, fn3_fba_runner, fn3_fba_ds, engine
    ):
        import dask

        dask.config.set(scheduler="threads")

        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, "test.h5")
            h = Harvester(fn3_fba_runner, fl_pth, engine=engine)
            h.harvest_cases([(1, 3), (2, 4)], chunks=1)
            h.harvest_cases([(1, 4), (2, 3)], chunks=1)
            assert not h.last_ds.identical(fn3_fba_ds)
            assert h.full_ds.identical(fn3_fba_ds)
            assert h.full_ds["sum"].chunks is not None
            h.full_ds.close()

    @pytest.mark.parametrize("engine", [netcdf4_case])
    def test_harvest_cases_merge_dask_default(
        self, fn3_fba_runner, fn3_fba_ds, engine
    ):
        import dask

        dask.config.set(scheduler="threads")

        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, "test.h5")
            h = Harvester(fn3_fba_runner, fl_pth, engine=engine, chunks=1)
            h.harvest_cases([(1, 3), (2, 4)])
            h.harvest_cases([(1, 4), (2, 3)])
            assert not h.last_ds.identical(fn3_fba_ds)
            assert h.full_ds.identical(fn3_fba_ds)
            assert h.full_ds["sum"].chunks is not None
            h.full_ds.close()

    def test_expand_dims_and_ellipsis_combos(self):
        @label("x", harvester=True)
        def foo(a, b, c=5):
            return a + b + c

        foo.harvest_combos(
            {
                "a": [1, 2],
                "b": [3, 4],
            }
        )

        assert foo.full_ds["x"].ndim == 2
        foo.expand_dims("c", 5)
        assert foo.full_ds["x"].ndim == 3

        foo.harvest_combos(
            {
                "a": ...,
                "b": ...,
                "c": [8, 9],
            }
        )
        assert list(foo.full_ds["c"].values) == [5, 8, 9]

    def test_drop_sel(self):
        @label("x", harvester=True)
        def foo(a, b, c):
            return a + b + c

        foo.harvest_combos(
            {
                "a": [1, 2],
                "b": [3, 4],
                "c": [5, 6, 7],
            }
        )

        assert foo.full_ds["x"].ndim == 3
        assert foo.full_ds["x"].size == 12
        foo.drop_sel(c=6)
        assert foo.full_ds["x"].ndim == 3
        assert foo.full_ds["x"].size == 8
        assert 6 not in foo.full_ds.coords["c"].values


class TestSampler:
    @pytest.mark.parametrize(
        "fname,engine",
        [
            ("test.pkl", "pickle"),
            ("test.csv", "csv"),
        ],
    )
    def test_sample_combos_new(self, fname, engine):
        @label(
            var_names=["sum", "diff", "divisor", "const"], constants={"c": 42}
        )
        def sum_diff(a, b, c):
            return a + b, a - b, a % b == 0, c

        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, fname)
            s = Sampler(sum_diff, fl_pth, engine=engine)
            print(s)
            combos = (
                ("a", (1, 2, 3, 4, 5)),
                ("b", lambda: np.random.randint(10, 20)),
            )
            s.sample_combos(10, combos)
            hdf = load_df(fl_pth, engine=engine)

            assert len(hdf) == 10
            assert s.last_df.compare(hdf).empty
            assert s.full_df.compare(hdf).empty

            s.sample_combos(20, combos)
            hdf = load_df(fl_pth, engine=engine)
            assert len(hdf) == 30
            assert not s.last_df.equals(hdf)
            assert s.full_df.equals(hdf)

            assert set(s.full_df["const"]) == set(s.full_df["c"]) == {42}

            for col in ["sum", "diff", "divisor", "const", "a", "b", "c"]:
                assert col in s.full_df

    @pytest.mark.parametrize("batchsize", [1, 3])
    def test_sow_reap_samples(self, batchsize):
        @label(
            var_names=["sum", "diff", "divisor", "const"], constants={"c": 42}
        )
        def sum_diff(a, b, c):
            return a + b, a - b, a % b == 0, c

        default_combos = (
            ("a", (1, 2, 3, 4, 5)),
            ("b", lambda: np.random.randint(10, 20)),
        )

        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, "test.pkl")
            s = Sampler(sum_diff, fl_pth, default_combos=default_combos)
            c = s.Crop("test-crop-def", batchsize=batchsize, parent_dir=tmpdir)
            c.sow_samples(10)
            c.grow_missing()
            c.reap()
            hdf = load_df(fl_pth)
            assert len(hdf) == 10
            assert s.last_df.compare(hdf).empty
            assert s.full_df.compare(hdf).empty



import collections

_MPL_MARKER_DICT = collections.OrderedDict(
    [
        ("circle", "o"),
        ("x", "x"),
        ("diamond", "D"),
        ("plus", "+"),
        ("square", "s"),
        ("point", "."),
        ("triangle_up", "^"),
        ("tri_left", "3"),
        ("triangle_right", ">"),
        ("thin_diamond", "d"),
        ("star", "*"),
        ("triangle_down", "v"),
        ("vline", "|"),
        ("tri_down", "1"),
        ("pentagon", "p"),
        ("pixel", ","),
        ("tri_up", "2"),
        ("triangle_left", "<"),
        ("hexagon1", "h"),
        ("tri_right", "4"),
        ("tickleft", 0),
        ("tickup", 2),
        ("tickdown", 3),
        ("caretleft", 4),
        ("hline", "_"),
        ("caretright", 5),
        ("hexagon2", "H"),
        ("tickright", 1),
        ("caretup", 6),
        ("octagon", "8"),
        ("caretdown", 7),
    ]
)
_MPL_MARKERS = [*_MPL_MARKER_DICT.values()]

_BOKEH_MARKER_DICT = collections.OrderedDict(
    [
        ("circle", "circle"),
        ("x", "x"),
        ("diamond", "diamond"),
        ("plus", "cross"),
        ("square", "square"),
        # ('point', '.'),
        ("triangle_up", "triangle"),
        # ('tri_left', '3'),
        # ('triangle_right', '>'),
        # ('thin_diamond', 'd'),
        ("star", "asterisk"),
        ("triangle_down", "inverted_triangle"),
        # ('vline', '|'),
        # ('tri_down', '1'),
        # ('pentagon', 'p'),
        # ('pixel', ','),
        # ('tri_up', '2'),
        # ('triangle_left', '<'),
        # ('hexagon1', 'h'),
        # ('tri_right', '4'),
        # ('tickleft', 0),
        # ('tickup', 2),
        # ('tickdown', 3),
        # ('caretleft', 4),
        # ('hline', '_'),
        # ('caretright', 5),
        # ('hexagon2', 'H'),
        # ('tickright', 1),
        # ('caretup', 6),
        # ('octagon', '8'),
        # ('caretdown', 7)
        ("circle_cross", "circle_cross"),
        ("square_x", "square_x"),
        ("diamond_cross", "diamond_cross"),
        ("circle_x", "circle_x"),
        ("square_cross", "square_cross"),
    ]
)
_BOKEH_MARKERS = [*_BOKEH_MARKER_DICT.values()]

_MARKERS = {"MATPLOTLIB": _MPL_MARKERS, "BOKEH": _BOKEH_MARKERS}
_SINGLE_LINE_MARKER = {"MATPLOTLIB": (".",), "BOKEH": ("circle",)}



import os
import tempfile

import numpy as np
import xarray as xr
from pytest import fixture, importorskip, mark, param, raises

from xyzpy.manage import (
    load_ds,
    save_ds,
    save_merge_ds,
)


@fixture
def ds1():
    return xr.Dataset(
        coords={"b": ["l1", "l2", "l4"], "a": [1, 2, 3, 4]},
        data_vars={
            "x": (
                ("b", "a"),
                np.random.randn(3, 4) + 1.0j * np.random.randn(3, 4),
            ),
            "isodd": ("a", np.asarray([True, False, True, False])),
        },
        attrs={"foo": "bar"},
    )


@fixture
def ds2():
    return xr.Dataset(
        coords={"b": ["l3", "l5"], "a": [3, 4, 5]},
        data_vars={
            "x": (
                ("b", "a"),
                np.random.randn(2, 3) + 1.0j * np.random.randn(2, 3),
            ),
            "isodd": ("a", np.asarray([True, False, True])),
        },
        attrs={"bar": "baz"},
    )


@fixture
def ds3():
    return xr.Dataset(
        coords={"b": ["l5"], "a": [4]},
        data_vars={
            "x": (("b", "a"), [[123.0 + 456.0j]]),
            "isodd": ("a", np.asarray([True])),
        },
        attrs={"baz": "qux"},
    )


@fixture
def ds_real():
    return xr.Dataset(
        coords={"b": ["l3", "l5"], "a": [3, 4, 5]},
        data_vars={
            "x": (("b", "a"), np.random.randn(2, 3)),
            "isodd": ("a", np.asarray([True, False, True])),
        },
        attrs={"qux": "corge"},
    )


class TestSaveAndLoad:
    @mark.parametrize(
        ("engine_save, engine_load"),
        [
            ("h5netcdf", "h5netcdf"),
            ("h5netcdf", "netcdf4"),
            ("zarr", "zarr"),
            ("joblib", "joblib"),
            param("netcdf4", "h5netcdf", marks=mark.xfail),
            param("netcdf4", "netcdf4", marks=mark.xfail),
        ],
    )
    def test_io_only_real(self, ds_real, engine_save, engine_load):
        importorskip(engine_save)
        importorskip(engine_load)
        with tempfile.TemporaryDirectory() as tmpdir:
            save_ds(
                ds_real, os.path.join(tmpdir, "test.h5"), engine=engine_save
            )
            ds2 = load_ds(os.path.join(tmpdir, "test.h5"), engine=engine_load)
            assert ds_real.equals(ds2)

    @mark.parametrize(
        ("engine_save, engine_load"),
        [
            ("h5netcdf", "h5netcdf"),
            ("zarr", "zarr"),
            ("joblib", "joblib"),
            param("h5netcdf", "netcdf4", marks=mark.xfail),
            param("netcdf4", "h5netcdf", marks=mark.xfail),
            param("netcdf4", "netcdf4", marks=mark.xfail),
        ],
    )
    def test_io_complex_data(self, ds1, engine_save, engine_load):
        importorskip(engine_save)
        importorskip(engine_load)

        with tempfile.TemporaryDirectory() as tmpdir:
            save_ds(ds1, os.path.join(tmpdir, "test.h5"), engine=engine_save)
            ds2 = load_ds(os.path.join(tmpdir, "test.h5"), engine=engine_load)
            assert ds1.identical(ds2)

    @mark.parametrize(("engine_load"), ["h5netcdf", "netcdf4"])
    def test_dask_load(self, ds_real, engine_load):
        importorskip(engine_load)

        with tempfile.TemporaryDirectory() as tmpdir:
            save_ds(ds_real, os.path.join(tmpdir, "test.nc"))
            ds2 = load_ds(
                os.path.join(tmpdir, "test.nc"), engine=engine_load, chunks=1
            )
            assert ds2.chunks["b"] == (1, 1)
            assert ds_real.identical(ds2)
            ds2.close()

    @mark.xfail
    def test_h5netcdf_dask(self, ds1):
        ds = ds1.chunk()
        with tempfile.TemporaryDirectory() as tmpdir:
            save_ds(ds, os.path.join(tmpdir, "test.h5"), engine="h5netcdf")
            ds2 = load_ds(os.path.join(tmpdir, "test.h5"))
            assert ds1.identical(ds2)
            ds2.close()

    def test_save_merge_ds(self, ds1, ds2, ds3):
        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, "test.h5")
            save_merge_ds(ds1, fname)
            save_merge_ds(ds2, fname)
            with raises(xr.MergeError):
                save_merge_ds(ds3, fname)
            save_merge_ds(ds3, fname, overwrite=True)
            exp = ds3.combine_first(xr.merge([ds1, ds2]))
            assert load_ds(fname).identical(exp)



import collections
import contextlib
import functools
import itertools
import warnings

import numpy as np

from .color import cmoke
from .plotter_matplotlib import (
    add_visualize_legend,
    show_and_close,
    to_colors,
)


@functools.lru_cache(16)
def get_neutral_style(draw_color=(0.5, 0.5, 0.5)):
    return {
        "axes.edgecolor": draw_color,
        "axes.facecolor": (0, 0, 0, 0),
        "axes.grid": True,
        "axes.labelcolor": draw_color,
        "axes.spines.right": False,
        "axes.spines.top": False,
        "figure.facecolor": (0, 0, 0, 0),
        "grid.alpha": 0.1,
        "grid.color": draw_color,
        "legend.frameon": False,
        "text.color": draw_color,
        "xtick.color": draw_color,
        "xtick.minor.visible": True,
        "ytick.color": draw_color,
        "ytick.minor.visible": True,
    }


def use_neutral_style(fn):
    """Decorator to use xyzpy neutral style for a function."""
    import matplotlib as mpl

    @functools.wraps(fn)
    def new_fn(
        *args, use_neutral_style=True, draw_color=(0.5, 0.5, 0.5), **kwargs
    ):
        if not use_neutral_style:
            return fn(*args, **kwargs)

        style = get_neutral_style(draw_color=draw_color)

        with mpl.rc_context(style):
            return fn(*args, **kwargs)

    return new_fn


@contextlib.contextmanager
def neutral_style(draw_color=(0.5, 0.5, 0.5), **kwargs):
    import matplotlib as mpl

    style = {
        **get_neutral_style(draw_color=draw_color),
        **kwargs,
    }

    with mpl.rc_context(style):
        yield


# colorblind palettes by Okabe & Ito: https://siegal.bio.nyu.edu/color-palette/

_COLORS_DEFAULT = (
    "#56B4E9",  # light blue
    "#E69F00",  # orange
    "#009E73",  # green
    "#D55E00",  # red
    "#F0E442",  # yellow
    "#CC79A7",  # purple
    "#0072B2",  # dark blue
)
_COLORS_SORTED = (
    "#0072B2",  # dark blue
    "#56B4E9",  # light blue
    "#009E73",  # green
    "#F0E442",  # yellow
    "#E69F00",  # orange
    "#D55E00",  # red
    "#CC79A7",  # purple
)


@functools.lru_cache(6)
def get_default_colormaps(N):
    return (
        cmoke(0.65, hue_shift=-0.05, val2=0.75),
        cmoke(0.13, hue_shift=-0.05, val2=0.80),
        cmoke(0.38, hue_shift=-0.05, val2=0.70),
        cmoke(0.04, hue_shift=-0.05, val2=0.60),
        cmoke(0.24, hue_shift=-0.05, val2=0.90),
        cmoke(0.90, hue_shift=-0.05, val2=0.75),
    )[:N]


def mod_sat(c, mod):
    """Modify the luminosity of rgb color ``c``."""
    from matplotlib.colors import hsv_to_rgb, rgb_to_hsv

    h, s, v = rgb_to_hsv(c[:3])
    return (*hsv_to_rgb((h, mod * s, v)), 1.0)


def auto_colors(N):
    import math

    from matplotlib.colors import LinearSegmentedColormap

    if N < len(_COLORS_DEFAULT):
        return _COLORS_DEFAULT[:N]

    cmap = LinearSegmentedColormap.from_list("okabe-ito", _COLORS_SORTED)

    xs = list(map(cmap, np.linspace(0, 1.0, N)))

    # modulate color saturation with sine to generate local distinguishability
    # ... but only turn on gradually for increasing number of nodes
    sat_mod_period = min(4, N / 7)
    sat_mod_factor = max(0.0, 2 / 3 * math.tanh((N - 7) / 4))

    return [
        mod_sat(
            c, 1 - sat_mod_factor * math.sin(math.pi * i / sat_mod_period) ** 2
        )
        for i, c in enumerate(xs)
    ]


def color_to_colormap(c, **autohue_opts):
    import matplotlib as mpl

    rgb = mpl.colors.to_rgb(c)
    h, s, v = mpl.colors.rgb_to_hsv(rgb)

    return cmoke(h, **autohue_opts)

    # vhi = min(1.0, v + vdiff / 2)
    # vlo = max(0.0, vhi - vdiff)
    # vhi = vlo + vdiff

    # shi = min(1.0, s + sdiff / 2)
    # slo = max(0.0, shi - sdiff)
    # shi = slo + sdiff

    # hsv_i = (h, max(slo, 0.0), min(vhi, 1.0))
    # hsv_f = (h, min(shi, 1.0), max(vlo, 0.0))

    # c1 = mpl.colors.hsv_to_rgb(hsv_i)
    # c2 = mpl.colors.hsv_to_rgb(hsv_f)
    # cdict = {
    #     "red": [(0.0, c1[0], c1[0]), (1.0, c2[0], c2[0])],
    #     "green": [(0.0, c1[1], c1[1]), (1.0, c2[1], c2[1])],
    #     "blue": [(0.0, c1[2], c1[2]), (1.0, c2[2], c2[2])],
    # }
    # return mpl.colors.LinearSegmentedColormap("", cdict)


def get_default_cmap(i, vdiff=0.5, sdiff=0.25):
    return color_to_colormap(_COLORS_DEFAULT[i], vdiff=vdiff, sdiff=sdiff)


def to_colormap(c, **autohue_opts):
    import numbers

    import matplotlib as mpl
    from matplotlib import pyplot as plt

    if isinstance(c, mpl.colors.Colormap):
        return c

    if isinstance(c, numbers.Number):
        return cmoke(c, **autohue_opts)

    try:
        return plt.get_cmap(c)
    except ValueError:
        return color_to_colormap(c, **autohue_opts)


def _make_bold(s):
    return r"$\bf{" + s.replace("_", r"\_") + r"}$"


_LINESTYLES_DEFAULT = (
    "solid",
    (0.0, (3, 1)),
    (0.5, (1, 1)),
    (1.0, (3, 1, 1, 1)),
    (1.5, (3, 1, 3, 1, 1, 1)),
    (2.0, (3, 1, 1, 1, 1, 1)),
)


_MARKERS_DEFAULT = (
    "o",
    "X",
    "v",
    "s",
    "P",
    "D",
    "^",
    "h",
    "*",
    "p",
    "<",
    "d",
    "8",
    ">",
    "H",
)

_HEATMAP_INVALID_KWARGS = {
    "hue",
    "color",
    "marker",
    "markersize",
    "markeredgecolor",
    "linewidth",
    "linestyle",
    "text",
}


_STYLE_SORT_ORDER = (
    "hue",
    "color",
    "markeredgecolor",
    "marker",
    "linestyle",
    "markersize",
    "linewidth",
)

INFINIPLOTTER_DEFAULTS = dict(
    bins=None,
    bins_density=False,
    aggregate=None,
    aggregate_method="median",
    aggregate_err_range=0.5,
    err=None,
    err_style=None,
    err_kws=None,
    xlink=None,
    color=None,
    colors=None,
    color_order=None,
    color_label=None,
    color_ticklabels=None,
    colormap_start=0.0,
    colormap_stop=1.0,
    hue=None,
    hues=None,
    hue_order=None,
    hue_label=None,
    hue_ticklabels=None,
    palette=None,
    autohue_start=0.25,
    autohue_sweep=1.0,
    autohue_opts=None,
    marker=None,
    markers=None,
    marker_order=None,
    marker_label=None,
    marker_ticklabels=None,
    markersize=None,
    markersizes=None,
    markersize_order=None,
    markersize_label=None,
    markersize_ticklabels=None,
    markeredgecolor="white",
    markeredgecolors=None,
    markeredgecolor_order=None,
    markeredgecolor_label=None,
    markeredgecolor_ticklabels=None,
    linewidth=None,
    linewidths=None,
    linewidth_order=None,
    linewidth_label=None,
    linewidth_ticklabels=None,
    linestyle=None,
    linestyles=None,
    linestyle_order=None,
    linestyle_label=None,
    linestyle_ticklabels=None,
    text=None,
    text_formatter=str,
    text_opts=None,
    col=None,
    col_order=None,
    col_label=None,
    col_ticklabels=None,
    row=None,
    row_order=None,
    row_label=None,
    row_ticklabels=None,
    alpha=1.0,
    join_across_missing=False,
    err_band_alpha=0.1,
    err_bar_capsize=1,
    xlabel=None,
    ylabel=None,
    xlim=None,
    ylim=None,
    zlim=None,
    xscale=None,
    yscale=None,
    zscale=None,
    xbase=10,
    ybase=10,
    xticks=None,
    yticks=None,
    xticklabels=None,
    yticklabels=None,
    vspans=(),
    hspans=(),
    span_color=(0.5, 0.5, 0.5),
    span_alpha=0.5,
    span_linewidth=1,
    span_linestyle=":",
    grid=True,
    grid_which="major",
    grid_alpha=0.1,
    legend=True,
    legend_ncol=None,
    legend_merge="auto",
    legend_reverse=False,
    legend_entries=None,
    legend_labels=None,
    legend_extras=None,
    legend_opts=None,
    label=None,
    title=None,
    ax=None,
    axs=None,
    format_axs=None,
    figsize=None,
    background_color=None,
    height=3,
    width=None,
    hspace=0.12,
    wspace=0.12,
    sharex=True,
    sharey=True,
    kwargs=None,
)

_PLOTTER_OPTS = tuple(INFINIPLOTTER_DEFAULTS.keys())


class Infiniplotter:
    def __init__(
        self,
        ds,
        x,
        y=None,
        z=None,
        **kwargs,
    ):
        import numpy as np
        from matplotlib import pyplot as plt

        self.x = x
        self.y = y
        self.z = z

        self.is_histogram = self.y is None
        self.is_heatmap = self.z is not None
        self.is_scatter = (not self.is_histogram) and (self.x in ds.data_vars)

        settings = {**INFINIPLOTTER_DEFAULTS, **kwargs}
        for opt in _PLOTTER_OPTS:
            setattr(self, opt, settings.pop(opt))

        # parse unmatched keywords and error but suggest correct versions
        if len(settings) > 0:
            import difflib

            wrong_opts = list(settings.keys())
            right_opts = [
                difflib.get_close_matches(opt, _PLOTTER_OPTS, n=3)
                for opt in wrong_opts
            ]

            msg = "Option(s) {} not valid.\n Did you mean: {}?".format(
                wrong_opts, right_opts
            )
            print(msg)
            raise ValueError(msg)

        # options to simply pass on to the matplotlib calls
        # use an explicit dict so we can autocorrect other options
        self.kwargs = self.kwargs or {}

        self.autohue_opts = (
            {} if self.autohue_opts is None else dict(self.autohue_opts)
        )

        if self.text is not None:
            self.text_opts = (
                {} if self.text_opts is None else dict(self.text_opts)
            )
            self.text_opts.setdefault("size", 6)
            self.text_opts.setdefault("horizontalalignment", "left")
            self.text_opts.setdefault("verticalalignment", "bottom")
            self.text_opts.setdefault("clip_on", True)

        self.err_kws = {} if self.err_kws is None else dict(self.err_kws)

        # if only one is specified allow it to be either
        if (self.hue is not None) and (self.color is None):
            for _a in ["", "_order", "s", "_label", "_ticklabels"]:
                setattr(self, f"color{_a}", getattr(self, f"hue{_a}"))
                setattr(self, f"hue{_a}", None)

        # default style options
        self.base_style = {
            "alpha": self.alpha,
            "markersize": 6,
            "color": "#0ca0eb",
            "marker": ".",
            "markeredgecolor": "white",
        }
        # the extent of each mapped dimension
        self.sizes = {}
        # the 'domain' of each mapped dimension (i.e. coordinate values)
        self.input_values = {}
        # the 'range' of each mapped dimension (e.g. marker or color)
        self.output_values = {}
        # how to label each mapped dimension
        self.labels = {
            self.x: self.x if self.xlabel is None else self.xlabel,
            self.y: self.y if self.ylabel is None else self.ylabel,
        }
        # how to name each point within each mapped dimension
        self.ticklabels = {}

        # work out all the dim mapping information

        def default_colormaps(N):
            if N <= 6:
                hs = get_default_colormaps(N)
            else:
                hs = np.linspace(
                    self.autohue_start,
                    self.autohue_start + self.autohue_sweep,
                    N,
                    endpoint=False,
                )

            self.autohue_opts.setdefault("hue_shift", 0.5 / N)
            return [to_colormap(h, **self.autohue_opts) for h in hs]

        # drop irrelevant data variables
        ds = ds.drop_vars(
            [
                k
                for k in ds
                if k
                not in (
                    self.x,
                    self.y,
                    self.z,
                    self.err,
                    self.hspans,
                    self.vspans,
                )
            ]
        )
        # then drop irrelevant dimensions
        possible_dims = set()
        if self.x in ds.data_vars:
            possible_dims.update(ds[self.x].dims)
        if self.y in ds.data_vars:
            possible_dims.update(ds[self.y].dims)
        if self.z in ds.data_vars:
            possible_dims.update(ds[self.z].dims)
        self.ds = ds.drop_dims([k for k in ds.dims if k not in possible_dims])

        self.mapped = set()

        self.init_mapped_dim(
            "hue",
            custom_values=(
                [to_colormap(h, **self.autohue_opts) for h in self.hues]
                if self.hues is not None
                else None
            ),
            default_values=default_colormaps,
        )
        self.init_mapped_dim(
            "color",
            custom_values=self.colors,
            default_values=lambda N: np.linspace(
                self.colormap_start, self.colormap_stop, N
            ),
        )

        if (self.hue is not None) and (self.color is not None):
            # need special label
            dim = f"{self.hue}, {self.color}"
            self.labels[dim] = (
                f"{self.labels[self.hue]}, {self.labels[self.color]}"
            )
            self.ticklabels[dim] = {}

        if (self.hue is None) and (self.color is not None):
            # set a global colormap or sequence
            if self.colors is None:
                self.cmap_or_colors = (
                    to_colormap(self.palette, **self.autohue_opts)
                    if self.palette is not None
                    else auto_colors(self.sizes["color"])
                )
            else:
                self.cmap_or_colors = self.output_values["color"]

        if self.is_scatter and (self.xlink is None):
            if self.linewidth is not None:
                warnings.warn(
                    "`linewidth` is not supported for scatters, ignoring. "
                    "Set `x` as a coordinate or use `xlink` for a line plot."
                )
            self.linewidth = 0
            if self.linestyle is not None:
                warnings.warn(
                    "`linestyle` is not supported for scatters, ignoring."
                    "Set `x` as a coordinate or use `xlink` for a line plot."
                )
            self.linestyle = ""

        self.init_mapped_dim(
            "markeredgecolor",
            custom_values=self.markeredgecolors,
            default_values=lambda N: auto_colors(N),
        )
        self.init_mapped_dim(
            "marker",
            custom_values=self.markers,
            default_values=itertools.cycle(_MARKERS_DEFAULT),
        )
        self.init_mapped_dim(
            "linestyle",
            custom_values=self.linestyles,
            default_values=itertools.cycle(_LINESTYLES_DEFAULT),
        )
        # XXX: set this automatically based on number of `x` values?
        self.init_mapped_dim(
            "markersize",
            custom_values=self.markersizes,
            default_values=lambda N: np.linspace(3.0, 9.0, N),
        )
        self.init_mapped_dim(
            "linewidth",
            custom_values=self.linewidths,
            default_values=lambda N: np.linspace(1.0, 3.0, N),
        )
        self.init_mapped_dim("col")
        self.init_mapped_dim("row")

        # compute which dimensions are not target or mapped dimensions
        self.unmapped = sorted(
            set(self.ds.dims)
            - self.mapped
            - {self.x, self.y, self.z, self.xlink}
        )

        if self.is_scatter and (self.xlink is None):
            # want to vectorize plotting over all unmapped dimensions
            self.ds = self.ds.stack({"__unmapped__": self.unmapped})

        if self.is_histogram:
            # histogram: create y as probability density / counts
            import xarray as xr

            # bin over all unmapped dimensions
            self.ds = self.ds.stack({"__unmapped__": self.unmapped})

            # work out the bin coordinates
            if self.bins is None or isinstance(self.bins, int):
                if self.bins is None:
                    nbins = min(
                        max(3, int(self.ds["__unmapped__"].size ** 0.5)), 50
                    )
                else:
                    nbins = self.bins
                xmin, xmax = self.ds[self.x].min(), self.ds[self.x].max()
                self.bins = np.linspace(float(xmin), float(xmax), nbins + 1)
            elif not isinstance(self.bins, np.ndarray):
                self.bins = np.asarray(self.bins)

            bin_coords = (self.bins[1:] + self.bins[:-1]) / 2

            if self.bins_density:
                self.y = f"prob({self.x})"
            else:
                self.y = f"count({self.x})"

            if self.ylabel is None:
                self.labels[self.y] = self.y
            else:
                self.labels[self.y] = self.ylabel

            ds_binned = (
                xr.apply_ufunc(
                    lambda x: np.histogram(
                        x, bins=self.bins, density=self.bins_density
                    )[0],
                    self.ds[self.x],
                    input_core_dims=[["__unmapped__"]],
                    output_core_dims=[[self.x]],
                    vectorize=True,
                )
                .to_dataset(name=self.y)
                .assign_coords({self.x: bin_coords})
            )

            if isinstance(self.hspans, str):
                ds_binned[self.hspans] = self.ds[self.hspans]
            if isinstance(self.vspans, str):
                ds_binned[self.vspans] = self.ds[self.vspans]

            self.ds = ds_binned
            self.kwargs.setdefault("drawstyle", "steps-mid")

        if self.is_heatmap and self.unmapped:
            if isinstance(self.aggregate, str):
                self.aggregate = [self.aggregate]

            if isinstance(self.aggregate, (list, tuple)) and set(
                self.aggregate
            ) == set(self.unmapped):
                # user has explicitly supplied all remaining dimensions
                self.aggregate = True

            elif self.aggregate is not True:
                # default to aggregating over all unmapped dimensions, but warn
                warnings.warn(
                    "Heatmap: aggregating over all unmapped dimensions: "
                    f"{self.unmapped}. Set `aggregate=True` to acknowledge "
                    "and disable this warning, or map the dimension(s) to "
                    "`row` or `col`."
                )
                self.aggregate = True

        # get the target data array and possibly aggregate some dimensions
        if self.aggregate:
            agg_into = self.z or self.y

            if self.aggregate is True:
                # select all unmapped dimensions
                self.aggregate = self.unmapped

            # compute data ranges to maybe show spread bars or bands
            if self.aggregate_err_range == "std":
                da_std_mean = self.ds[agg_into].mean(self.aggregate)
                da_std = self.ds[agg_into].std(self.aggregate)

                self.da_ql = da_std_mean - da_std
                self.da_qu = da_std_mean + da_std

            elif self.aggregate_err_range == "stderr":
                da_stderr_mean = self.ds[agg_into].mean(self.aggregate)
                da_stderr_cnt = self.ds[agg_into].notnull().sum(self.aggregate)
                da_stderr = self.ds[agg_into].std(self.aggregate) / np.sqrt(
                    da_stderr_cnt
                )

                self.da_ql = da_stderr_mean - da_stderr
                self.da_qu = da_stderr_mean + da_stderr

            else:
                self.aggregate_err_range = min(
                    max(0.0, self.aggregate_err_range), 1.0
                )
                ql = 0.5 - self.aggregate_err_range / 2.0
                qu = 0.5 + self.aggregate_err_range / 2.0
                self.da_ql = self.ds[agg_into].quantile(ql, self.aggregate)
                self.da_qu = self.ds[agg_into].quantile(qu, self.aggregate)

            # default to showing spread as bands
            if self.err is None:
                self.err = True
            if self.err_style is None:
                self.err_style = "band"

            # main data for central line
            self.ds = getattr(self.ds, self.aggregate_method)(self.aggregate)

        # default to bars if err not taken from aggregating
        if self.err_style is None:
            self.err_style = "bars"

        # all the coordinates we will iterate over
        self.remaining_dims = sorted(
            # only iterate over dimensions not forming the actual line etc
            (
                dim
                for dim in self.ds.dims
                if dim
                not in (self.x, self.y, self.z, self.xlink, "__unmapped__")
            ),
            # for legend: we want to iterate over hue first, then color, etc.
            key=lambda dim: tuple(
                dim != getattr(self, name, None) for name in _STYLE_SORT_ORDER
            ),
        )
        self.remaining_ranges = [
            range(self.ds.sizes[dim]) for dim in self.remaining_dims
        ]

        # finally, maybe create the figure and axes
        if self.ax is not None:
            if self.axs is not None:
                raise ValueError("cannot specify both `ax` and `axs`")
            self.axs = np.array([[self.ax]])

        if self.axs is None:
            if self.figsize is None:
                if self.width is None:
                    self.width = self.height
                if self.height is None:
                    self.height = self.width
                self.figsize = (
                    self.width * self.sizes["col"],
                    self.height * self.sizes["row"],
                )

            self.fig, self.axs = plt.subplots(
                self.sizes["row"],
                self.sizes["col"],
                sharex=self.sharex,
                sharey=self.sharey,
                squeeze=False,
                gridspec_kw={"hspace": self.hspace, "wspace": self.wspace},
                figsize=self.figsize,
            )
            if self.background_color is not None:
                self.fig.patch.set_facecolor(self.background_color)
            else:
                self.fig.patch.set_alpha(0.0)
        else:
            self.fig = None

        if (self.fig is not None) and (self.title is not None):
            self.fig.suptitle(self.title)

    def init_mapped_dim(
        self,
        name,
        custom_values=None,
        default_values=None,
    ):
        dim = getattr(self, name)

        if isinstance(dim, list):
            # make hashable
            dim = tuple(dim)

        # handle fused dimensions
        if isinstance(dim, tuple):
            # fused named
            new_dim = ", ".join(dim)
            if new_dim in self.ds.dims:
                # already fused -> nothing to do
                dim = new_dim
            elif all(x in self.ds.dims for x in dim):
                # create a new fused dimension
                self.ds = self.ds.stack({new_dim: dim})
                dim = new_dim
            # else not a valid fused dimensions -> assume constant property

        if (dim is not None) and (dim not in self.ds.dims):
            # attribute is just manually specified, not mapped to dimension
            self.base_style[name] = dim
            self.sizes[name] = 1
            setattr(self, name, None)
            return

        order = getattr(self, f"{name}_order")
        dim_label = getattr(self, f"{name}_label")
        dim_ticklabels = getattr(self, f"{name}_ticklabels")

        if dim is not None:
            if self.is_heatmap and (name in _HEATMAP_INVALID_KWARGS):
                raise ValueError(f"Heatmap: cannot map property `{name}`.")

            if order is not None:
                # select and order along dimension
                self.ds = self.ds.sel({dim: list(order)})

            self.mapped.add(dim)
            self.ds = self.ds.dropna(dim, how="all")

            self.input_values[name] = self.ds[dim].values
            self.sizes[name] = len(self.input_values[name])
            self.labels[dim] = (
                _make_bold(dim) if dim_label is None else dim_label
            )

            if dim_ticklabels is None:
                dim_ticklabels = {}
            if not isinstance(dim_ticklabels, dict):
                dim_ticklabels = dict(
                    zip(self.input_values[name], dim_ticklabels)
                )
            self.ticklabels[dim] = dim_ticklabels

            if custom_values is None:
                if default_values is not None:
                    if callable(default_values):
                        # allow default values to depend on number of values
                        default_values = default_values(self.sizes[name])

                    self.output_values[name] = tuple(
                        x
                        for x, _ in zip(
                            default_values, range(self.sizes[name])
                        )
                    )
            else:
                self.output_values[name] = custom_values
        else:
            self.sizes[name] = 1

        setattr(self, name, dim)

    def plot_lines(self):
        """Plot lines of the data."""
        import matplotlib as mpl

        # line handles for merged legend
        #     keys: frozenset([(mapped_dim, dim_value), ...]) ->
        #     values: Line2D handle with respective style
        self.handles_merged = {}

        # line handles for split legend, nested dict:
        #    keys: mapped_dim ->
        #    values: dict with
        #        keys: dim_value ->
        #        values: dict with
        #            keys: style property (e.g. color, marker, linestyle) ->
        #            values: property value (e.g. 'red', 'o', '--')
        self.handles_split = collections.defaultdict(
            lambda: collections.defaultdict(dict)
        )

        x_is_constant = self.x not in self.ds.data_vars
        if x_is_constant:
            # is a constant coordinate
            xdata = self.ds[self.x].values

        # iterate over and plot all data
        for iloc in itertools.product(*self.remaining_ranges):
            # current coordinates
            loc = dict(zip(self.remaining_dims, iloc))

            # get the correct axes to plot on
            if self.row is not None:
                i_ax = loc[self.row]
            else:
                i_ax = 0
            if self.col is not None:
                j_ax = loc[self.col]
            else:
                j_ax = 0
            ax = self.axs[i_ax, j_ax]

            # map coordinate into relevant styles and track each uniquely
            sub_key = {}
            # style options specific to this trace, to be supplied to ax.plot
            specific_style = {}

            # need to handle hue and color separately
            if self.color is not None:
                if self.hue is not None:
                    ihue = loc[self.hue]
                    hue_in = self.input_values["hue"][ihue]
                    sub_key[self.hue] = hue_in
                    self.cmap_or_colors = self.output_values["hue"][ihue]

                icolor = loc[self.color]
                color_in = self.input_values["color"][icolor]
                if not callable(self.cmap_or_colors):
                    color_out = self.cmap_or_colors[icolor]
                else:
                    color_out = self.cmap_or_colors(
                        self.output_values["color"][icolor]
                    )

                sub_key[self.color] = color_in
                specific_style["color"] = color_out
                if self.hue is None:
                    legend_dim = self.color
                    legend_in = color_in
                else:
                    legend_dim = ", ".join((self.hue, self.color))
                    legend_in = ", ".join(map(str, (hue_in, color_in)))

                self.handles_split[legend_dim][legend_in]["color"] = color_out
            else:
                legend_dim = None

            for prop in _STYLE_SORT_ORDER:
                if prop in ("color", "hue"):
                    # already handled
                    continue

                dim = getattr(self, prop)
                if dim is not None:
                    idx = loc[dim]
                    prop_in = self.input_values[prop][idx]
                    prop_out = self.output_values[prop][idx]
                    sub_key[dim] = prop_in
                    specific_style[prop] = prop_out

                    if dim in (self.color, self.hue):
                        self.handles_split[legend_dim][legend_in][prop] = (
                            prop_out
                        )
                    else:
                        self.handles_split[dim][prop_in][prop] = prop_out

            # get the masked x and y data
            ds_loc = self.ds.isel(loc)
            mask = ds_loc[self.y].notnull().values

            if not x_is_constant:
                # x also varying
                xdata = ds_loc[self.x].values
                # both x and y must be non-null
                mask &= ds_loc[self.x].notnull().values

            if not np.any(mask):
                # don't plot all null lines
                continue

            if not self.join_across_missing:
                # reset mask
                data_mask = ()
            else:
                data_mask = mask

            xmdata = xdata[data_mask]
            ymdata = ds_loc[self.y].values[data_mask]

            if self.err is not None:
                if (self.err is True) and (self.aggregate is not None):
                    da_ql_loc = self.da_ql.isel(loc)
                    da_qu_loc = self.da_qu.isel(loc)
                    y1 = da_ql_loc.values[data_mask]
                    y2 = da_qu_loc.values[data_mask]
                    yneg = ymdata - y1
                    ypos = y2 - ymdata
                else:
                    yerr_mdata = ds_loc[self.err].values[data_mask]
                    yneg = -yerr_mdata
                    ypos = +yerr_mdata
                    y1 = ymdata + yneg
                    y2 = ymdata + ypos

                if self.err_style == "bars":
                    ax.errorbar(
                        x=xmdata,
                        y=ymdata,
                        yerr=[np.abs(yneg), np.abs(ypos)],
                        fmt="none",
                        capsize=self.err_bar_capsize,
                        **{
                            **self.base_style,
                            **specific_style,
                            **self.err_kws,
                        },
                    )
                elif self.err_style == "band":
                    ax.fill_between(
                        x=xmdata,
                        y1=y1,
                        y2=y2,
                        color=specific_style.get(
                            "color", self.base_style["color"]
                        ),
                        alpha=self.err_band_alpha,
                        **self.err_kws,
                    )

            if self.is_histogram:
                ax.fill_between(
                    x=xmdata,
                    y1=ymdata,
                    y2=0,
                    step={
                        None: None,
                        "default": None,
                        "steps": "pre",
                        "steps-pre": "pre",
                        "steps-mid": "mid",
                        "steps-post": "post",
                    }[self.kwargs.get("drawstyle", None)],
                    color=mpl.colors.to_rgb(
                        specific_style.get("color", self.base_style["color"])
                    ),
                    alpha=self.err_band_alpha,
                )

            plot_opts = {**self.base_style, **specific_style}

            if self.label:
                # override with manual label
                label = self.label
            else:
                # label = ", ".join(map(str, sub_key.values()))
                label = ", ".join(
                    self.ticklabels[dim].get(val, str(val))
                    for dim, val in sub_key.items()
                )

            # do the plotting!
            (handle,) = ax.plot(
                xmdata,
                ymdata,
                label=label,
                **plot_opts,
                **self.kwargs,
            )

            # add a text label next to each point
            if self.text is not None:
                # need raw mask for text
                smdata = ds_loc[self.text].values[mask]
                for txx, txy, txs in zip(
                    xdata[mask], ds_loc[self.y].values[mask], smdata
                ):
                    specific_text_opts = {}
                    if "color" not in self.text_opts:
                        # default to line color
                        specific_text_opts["color"] = plot_opts["color"]

                    ax.text(
                        txx,
                        txy,
                        self.text_formatter(txs),
                        **self.text_opts,
                        **specific_text_opts,
                    )

            # only want one legend entry per unique style
            key = frozenset(sub_key.items())
            if key or self.label:
                # has a mapped style (or a manually specified label)
                self.handles_merged.setdefault(key, handle)

            # add spans that depend on location
            for _spans, ax_func in (
                (self.hspans, ax.axhline),
                (self.vspans, ax.axvline),
            ):
                if isinstance(_spans, str):
                    for spn in ds_loc[_spans].values.flat:
                        ax_func(
                            spn,
                            color=self.span_color,
                            alpha=self.span_alpha,
                            linestyle=self.span_linestyle,
                            linewidth=self.span_linewidth,
                        )

        self.do_axes_formatting()

        if self.legend:
            self.create_legend()

    def create_legend(self):
        """Create a legend for lines plot."""
        from matplotlib.lines import Line2D

        legend_opts = {} if self.legend_opts is None else self.legend_opts

        if ("bbox_to_anchor" in legend_opts) or ("loc" in legend_opts):
            # we are manually placing legend, assume its a new one, rather
            # than trying to extend
            try:
                legend_old = self.axs[0, -1].get_legend()
                # later call to legend will overwrite this one: explicitly add
                self.axs[0, -1].add_artist(legend_old)
            except AttributeError:
                pass
            legend_handles = []

        else:
            # try to extend current legend with more entries
            try:
                legend_handles = self.axs[0, -1].get_legend().get_lines()
                # add a space
                legend_handles.append(
                    Line2D([0], [0], markersize=0, linewidth=0, label="")
                )
            except AttributeError:
                legend_handles = []

        if self.legend_merge == "auto":
            self.legend_merge = (
                # all mapped style correspond to same dimension
                (len(self.mapped) <= 1)
                # color and hue are always effectively fused
                or (set(self.mapped) == {self.hue, self.color})
            )

        if self.handles_merged and self.legend_merge:
            # entry for every unique style combination
            self._create_legend_merged(legend_handles, legend_opts)
        elif self.handles_split:
            # separate legend for each style separately
            self._create_legend_split(legend_handles, legend_opts)
        else:
            legend_handles = None

        if legend_handles is not None:
            # we have some legend entries

            if self.legend_labels is not None:
                # we have explicit labels for each handle
                for lh, label in zip(
                    # only update the handles that were added by this plot call
                    legend_handles[-len(self.handles_merged) :],
                    self.legend_labels,
                ):
                    lh.set_label(label)

            if (
                legend_opts.get("title", "") == "" and len(legend_handles) == 1
            ) and (self.label is None):
                # promote single legend entry to title
                lh0 = legend_handles[0]
                legend_opts["title"] = lh0.get_label()
                lh0.set_label("")

            lax = self.axs[0, -1]
            legend_opts.setdefault("loc", "upper left")
            legend_opts.setdefault("bbox_to_anchor", (1.0, 1.0))
            legend_opts.setdefault("columnspacing", 1.0)
            legend_opts.setdefault("edgecolor", "none")
            legend_opts.setdefault("framealpha", 0.0)
            lax.legend(handles=legend_handles, **legend_opts)

    def _create_legend_merged(self, legend_handles, legend_opts):
        """Entry for every unique style combination."""

        if self.legend_entries:
            # only keep manually specified legend entries
            remove = set()
            for k in self.handles_merged:
                for dim, val in k:
                    if dim in self.legend_entries:
                        if val not in self.legend_entries[dim]:
                            remove.add(k)
            for k in remove:
                del self.handles_merged[k]

        # note: we explicitly iterated over the data is desired legend order
        handles = list(self.handles_merged.values())
        if self.legend_reverse:
            handles.reverse()
        legend_handles.extend(handles)

        if self.legend_ncol is None:
            if self.sizes["color"] == 1 or len(self.handles_merged) <= 10:
                self.legend_ncol = 1
            else:
                self.legend_ncol = self.sizes["hue"]

        legend_title = []
        for prop in _STYLE_SORT_ORDER:
            dim = getattr(self, prop)
            if dim is not None and self.labels[dim] not in legend_title:
                # check if not in legend_title, as multiple style
                # properties can be mapped to the same dimension
                legend_title.append(self.labels[dim])

        legend_opts.setdefault("title", ", ".join(legend_title))
        legend_opts.setdefault("ncol", self.legend_ncol)

    def _create_legend_split(self, legend_handles, legend_opts):
        """Separate legend for each style separately."""
        from matplotlib.lines import Line2D

        if self.legend_entries:
            # only keep manually specified legend entries
            for k, vals in self.legend_entries.items():
                self.handles_split[k] = {
                    key: val
                    for key, val in self.handles_split[k].items()
                    if key in vals
                }

        self.base_style["color"] = (0.5, 0.5, 0.5)
        self.base_style["marker"] = ""
        self.base_style["linestyle"] = ""

        ncol = len(self.handles_split)
        nrow = max(map(len, self.handles_split.values()))

        for legend_dim, dimval_to_style in self.handles_split.items():
            # add sub title in legend
            legend_handles.append(
                Line2D(
                    [0],
                    [0],
                    markersize=0,
                    linewidth=0,
                    label=self.labels[legend_dim],
                )
            )

            dimval_styles = list(dimval_to_style.items())
            if self.legend_reverse:
                dimval_styles.reverse()

            for dimval, style in dimval_styles:
                # XXX: maybe pretty print float values etc here?
                label = self.ticklabels[legend_dim].get(dimval, str(dimval))

                if any("marker" in prop for prop in style):
                    style.setdefault("marker", "o")
                if any("line" in prop for prop in style):
                    style.setdefault("linestyle", "-")
                if "color" in style:
                    style.setdefault("marker", ".")
                    style.setdefault("linestyle", "-")

                legend_handles.append(
                    Line2D(
                        [0],
                        [0],
                        **{**self.base_style, **style},
                        label=label,
                    )
                )

            if self.legend_ncol is None:
                npad = nrow - len(dimval_styles)
            else:
                npad = 1
            for _ in range(npad):
                # spacing
                legend_handles.append(
                    Line2D([0], [0], markersize=0, linewidth=0, label="")
                )

        if self.legend_ncol is None:
            legend_opts.setdefault("ncol", ncol)

        if self.legend_extras is not None:
            for extra in self.legend_extras:
                if isinstance(extra, dict):
                    extra = Line2D([0], [0], **extra)
                legend_handles.append(extra)

    def plot_heatmap(self):
        """Plot a heatmap of the data."""
        import matplotlib as mpl

        xdata = self.ds[self.x].values
        ydata = self.ds[self.y].values

        zdata_all = self.ds[self.z].values
        zdata_all = zdata_all[np.isfinite(zdata_all)]

        # set coloring limits
        try:
            zmin, zmax = self.zlim
        except (TypeError, ValueError):
            zmin = zmax = None
        if zmax is None:
            zmax = np.max(zdata_all)
        if zmin is None:
            zmin = np.min(zdata_all)
        max_mag = max(abs(zmax), abs(zmin))

        if self.palette is None:
            self.norm = None
        elif self.zscale == "log":
            self.norm = mpl.colors.LogNorm(vmin=zmin, vmax=zmax)
        elif self.zscale == "symlog":
            self.norm = mpl.colors.SymLogNorm(
                vmin=zmin, vmax=zmax, linthresh=1
            )
        else:
            self.norm = mpl.colors.Normalize(vmin=zmin, vmax=zmax)

        for iloc in itertools.product(*self.remaining_ranges):
            # current coordinates
            loc = dict(zip(self.remaining_dims, iloc))

            # get the right set of axes to plot on
            if self.row is not None:
                i_ax = loc[self.row]
            else:
                i_ax = 0
            if self.col is not None:
                j_ax = loc[self.col]
            else:
                j_ax = 0
            ax = self.axs[i_ax, j_ax]

            zdata = self.ds[self.z].isel(loc).transpose(self.y, self.x).values

            # get the masked x and y data
            if self.palette is None:
                C = zdata
                mask = np.isfinite(C)
                zdata = np.empty(C.shape + (4,))
                zdata[mask] = to_colors(
                    C[mask], alpha_pow=0.0, max_mag=max_mag
                )[0]
                zdata[~mask] = (0.5, 0.5, 0.5, 0.5)
            else:
                zdata = (
                    self.ds[self.z].isel(loc).transpose(self.y, self.x).values
                )

            ax.pcolormesh(
                xdata,
                ydata,
                zdata,
                shading="nearest",
                # edgecolors="face",
                rasterized=True,
                cmap=self.palette,
                norm=self.norm,
            )

        self.do_axes_formatting()

        if self.legend:
            if self.palette is None:
                add_visualize_legend(
                    ax=self.axs[0, -1],
                    complexobj=np.iscomplexobj(zdata_all),
                    max_mag=max_mag,
                    legend_loc=(1.1, 0.8),
                    legend_size=0.2,
                    # max_projections=max_projections,
                    # legend_bounds=legend_bounds,
                    # legend_resolution=legend_resolution,
                )
            else:
                cax = self.axs[0, -1].inset_axes((1.1, 0.1, 0.05, 0.8))

                self.fig.colorbar(
                    mpl.cm.ScalarMappable(norm=self.norm, cmap=self.palette),
                    cax=cax,
                    orientation="vertical",
                    label=_make_bold(self.z),
                )

    def do_axes_formatting(self):
        if (self.fig is None) and (not self.format_axs):
            return

        # perform axes level formatting
        from matplotlib.ticker import (
            AutoMinorLocator,
            LogLocator,
            NullFormatter,
            ScalarFormatter,
            StrMethodFormatter,
        )

        for (i, j), ax in np.ndenumerate(self.axs):
            # only change this stuff if we created the figure
            title = []
            if self.col is not None:
                title.append(
                    f"{self.labels[self.col]}={self.input_values['col'][j]}"
                )
            if self.row is not None:
                title.append(
                    f"{self.labels[self.row]}={self.input_values['row'][i]}"
                )
            if title:
                title = ", ".join(title)
                ax.text(
                    0.5,
                    1.0,
                    title,
                    transform=ax.transAxes,
                    horizontalalignment="center",
                    verticalalignment="bottom",
                )

            # only label outermost plot axes
            if i + 1 == self.sizes["row"]:
                ax.set_xlabel(self.labels[self.x])
            if j == 0:
                ax.set_ylabel(self.labels[self.y])

            # set some nice defaults
            ax.spines["right"].set_visible(False)
            ax.spines["top"].set_visible(False)

            if self.grid:
                ax.grid(True, which=self.grid_which, alpha=self.grid_alpha)
                ax.set_axisbelow(True)

            if self.xlim is not None:
                ax.set_xlim(self.xlim)
            if self.ylim is not None:
                ax.set_ylim(self.ylim)

            if self.xscale is not None:
                ax.set_xscale(self.xscale)
            if self.yscale is not None:
                ax.set_yscale(self.yscale)

            for scale, base, ticks, ticklabels, axis in [
                (
                    self.xscale,
                    self.xbase,
                    self.xticks,
                    self.xticklabels,
                    ax.xaxis,
                ),
                (
                    self.yscale,
                    self.ybase,
                    self.yticks,
                    self.yticklabels,
                    ax.yaxis,
                ),
            ]:
                if scale == "log":
                    axis.set_major_locator(LogLocator(base=base, numticks=6))
                    if base != 10:
                        if isinstance(base, int):
                            axis.set_major_formatter(
                                StrMethodFormatter("{x:.0f}")
                            )
                        else:
                            axis.set_major_formatter(ScalarFormatter())
                    if base < 3:
                        subs = [1.5]
                    else:
                        subs = np.arange(2, base)
                    axis.set_minor_locator(LogLocator(base=base, subs=subs))
                    axis.set_minor_formatter(NullFormatter())
                elif scale == "symlog":
                    # TODO: choose some nice defaults
                    pass
                else:
                    axis.set_minor_locator(AutoMinorLocator(5))

                if ticks is not None:
                    axis.set_ticks(ticks, labels=ticklabels)

            for _spans, ax_func, h_or_v in (
                (self.hspans, ax.axhline, "h"),
                (self.vspans, ax.axvline, "v"),
            ):
                if not isinstance(_spans, str):
                    for line in _spans:
                        if isinstance(_spans, dict):
                            import matplotlib.transforms as transforms

                            # label each span using key from dict
                            spanlabel = line
                            # actual line location is value in dict
                            line = _spans[spanlabel]

                            if h_or_v == "h":
                                trans = transforms.blended_transform_factory(
                                    ax.transAxes, ax.transData
                                )
                                _span_label_coo = (0.0, line)
                                va = "bottom"
                            else:
                                trans = transforms.blended_transform_factory(
                                    ax.transData, ax.transAxes
                                )
                                _span_label_coo = (line, 0.0)
                                va = "bottom"

                            ax.text(
                                *_span_label_coo,
                                spanlabel,
                                ha="left",
                                va=va,
                                transform=trans,
                                color=self.span_color,
                            )

                        ax_func(
                            line,
                            color=self.span_color,
                            alpha=self.span_alpha,
                            linestyle=self.span_linestyle,
                            linewidth=self.span_linewidth,
                        )


@show_and_close
@use_neutral_style
def infiniplot(
    ds,
    x,
    y=None,
    z=None,
    **kwargs,
):
    """Helper class for the infiniplot functionality.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to plot.
    x : str
        Name of the x coordinate.
    y : str, optional
        Name of the y coordinate. If not specified, histogram mode is activated
        and the values of ``x`` are binned to produce a density or frequency to
        use as the y-variable.
    z : str, optional
        Name of the z coordinate. If specified this turns on the heatmap mode.
    bins : int or array_like, optional
        If in histogram mode, specify either the number of bins to use or the
        bin edges. If not specified, a default number of bins is automatically
        chosen based on the number of data points.
    bins_density : bool, optional
        If in histogram mode, whether to plot the density (True) or frequency
        (False) of the data. Default is True.
    aggregate : str or Sequence[str], optional
        If specified, aggregate over the given dimension(s) using
        ``aggregate_method`` (by default 'median'). If `True` aggregate over
        all unmapped dimensions. If in heatmap mode, this is automatically set
        to `True`, since only one plot can be shown per axis.
    aggregate_method : str, optional
        If ``aggregate`` is specified, the method to use for aggregation. Any
        option available as a method on a DataArray can be used, e.g. 'mean',
        'median', 'max'. Default is 'median'.
    aggregate_err_range : float or str, optional
        If ``aggregate`` is specified, the range of the error bars or bands to
        show. The options are:

        - ``'std'``: show the standard deviation of the data
        - ``'stderr'``: show the standard error of the mean
        - float: show the given quantile range, e.g. 0.5 for the interquartile
            range

    err : str, optional
        If specified, a data variable to use for error bars or bands. This
        overrides any derived from ``aggregate``.
    err_style : str, optional
        If specified, the style of error to show. The options are:

        - ``'bars'``: show error bars
        - ``'band'``: show error bands

    err_kws : dict, optional
        Additional keyword arguments to pass to the error plotting function.
    xlink : str, optional
        If specified, the name of a dimension to use for linking the x-axis.
        Used when you are plotting a variable rather than coordinate as ``x``,
        but want to link each sweep of values as a line.
    color : str, optional
        If specified, the name of a dimension to use for mapping the color or
        intensity of each line. If ``hue`` is also specified, this controls the
        intensity of the color. If not a dimension, this is used as a constant
        color for all lines.
    colors : sequence, optional
        An explicit sequence of colors to use for the color-mapped dimension.
    color_order : sequence, optional
        An explicit order of values to use for the color-mapped dimension.
    color_label : str, optional
        An alternate label to use for the color-mapped dimension.
    color_ticklabels : dict or sequence, optional
        A mapping from values to tick labels to use for the color-mapped
        dimension.
    colormap_start : float, optional
        If using a palette, the starting value of the colormap to use, e.g. 0.2
        would skip the first 20% of the colormap.
    colormap_stop : float, optional
        If using a palette, the stopping value of the colormap to use, e.g. 0.9
        would skip the last 10% of the colormap.
    hue : str, optional
        If specified, the name of a dimension to use for mapping the color or
        hue of each line. If ``color`` is also specified, this controls the hue
        of the color. If not a dimension, this is used as a constant hue for
        all lines.
    hues : sequence, optional
        An explicit sequence of hues to use for the hue-mapped dimension.
    hue_order : sequence, optional
        An explicit order of values to use for the hue-mapped dimension.
    hue_label : str, optional
        An alternate label to use for the hue-mapped dimension.
    hue_ticklabels : dict or sequence, optional
        A mapping from values to tick labels to use for the hue-mapped
        dimension.
    palette : str, sequence, or colormap, optional
        If specified, the name of a colormap, or an actual colormap, to use for
        mapping the color or hue of each line. If both ``color`` and ``hue``
        are specified, you can supply a sequence of palettes here, with ``hue``
        controlling which palette, and ``color`` controlling the intensity
        within the palette.
    autohue_start : float, optional
        If not using a palette, the starting hue to use for automatically
        generating a sequence of hues.
    autohue_sweep : float, optional
        If not using a palette, the sweep of hues to use for automatically
        generating a sequence of hues.
    autohue_opts : dict, optional
        Additional keyword arguments to pass to the automatic hue generator -
        see {func}`xyzpy.color.cmoke`.
    marker : str, optional
        If specified, the name of a dimension to use for mapping the marker
        style of each line. If not a dimension, this is used as a constant
        marker style for all lines.
    markers : sequence, optional
        An explicit sequence of markers to use for the marker-mapped dimension.
    marker_order : sequence, optional
        An explicit order of values to use for the marker-mapped dimension.
    marker_label : str, optional
        An alternate label to use for the marker-mapped dimension.
    marker_ticklabels : dict or sequence, optional
        A mapping from values to tick labels to use for the marker-mapped
        dimension.
    markersize : str, optional
        If specified, the name of a dimension to use for mapping the marker
        size of each line. If not a dimension, this is used as a constant
        marker size for all lines.
    markersizes : sequence, optional
        An explicit sequence of marker sizes to use for the markersize-mapped
        dimension.
    markersize_order : sequence, optional
        An explicit order of values to use for the markersize-mapped dimension.
    markersize_label : str, optional
        An alternate label to use for the markersize-mapped dimension.
    markersize_ticklabels : dict or sequence, optional
        A mapping from values to tick labels to use for the markersize-mapped
        dimension.
    markeredgecolor : str, optional
        If specified, the name of a dimension to use for mapping the marker
        edge color of each line. If not a dimension, this is used as a constant
        marker edge color for all lines.
    markeredgecolors : sequence, optional
        An explicit sequence of marker edge colors to use for the
        markeredgecolor-mapped dimension.
    markeredgecolor_order : sequence, optional
        An explicit order of values to use for the markeredgecolor-mapped
        dimension.
    markeredgecolor_label : str, optional
        An alternate label to use for the markeredgecolor-mapped dimension.
    markeredgecolor_ticklabels : dict or sequence, optional
        A mapping from values to tick labels to use for the
        markeredgecolor-mapped dimension.
    linewidth : str, optional
        If specified, the name of a dimension to use for mapping the line
        width of each line. If not a dimension, this is used as a constant
        line width for all lines.
    linewidths : sequence, optional
        An explicit sequence of line widths to use for the linewidth-mapped
        dimension.
    linewidth_order : sequence, optional
        An explicit order of values to use for the linewidth-mapped dimension.
    linewidth_label : str, optional
        An alternate label to use for the linewidth-mapped dimension.
    linewidth_ticklabels :
        A mapping from values to tick labels to use for the linewidth-mapped
        dimension.
    linestyle : str, optional
        If specified, the name of a dimension to use for mapping the line
        style of each line. If not a dimension, this is used as a constant
        line style for all lines.
    linestyles : sequence, optional
        An explicit sequence of line styles to use for the linestyle-mapped
        dimension.
    linestyle_order : sequence, optional
        An explicit order of values to use for the linestyle-mapped dimension.
    linestyle_label : str, optional
        An alternate label to use for the linestyle-mapped dimension.
    linestyle_ticklabels : dict or sequence, optional
        A mapping from values to tick labels to use for the linestyle-mapped
        dimension.
    text : str, optional
        If specified, the name of a dimension to use for mapping text
        annotations to each line.
    text_formatter : callable, optional
        A function to use to format data entries to text annotations. Default
        is ``str``.
    text_opts : dict, optional
        Additional keyword arguments to pass to the text plotting function.
    col : str, optional
        If specified, the name of a dimension to use for mapping the subplot
        column of each line.
    col_order : sequence, optional
        An explicit order of values to use for the col-mapped dimension.
    col_label : str, optional
        An alternate label to use for the col-mapped dimension.
    col_ticklabels : dict or sequence, optional
        A mapping from values to tick labels to use for the col-mapped
        dimension.
    row : str, optional
        If specified, the name of a dimension to use for mapping the subplot
        row of each line.
    row_order : sequence, optional
        An explicit order of values to use for the row-mapped dimension.
    row_label : str, optional
        An alternate label to use for the row-mapped dimension.
    row_ticklabels : dict or sequence, optional
        A mapping from values to tick labels to use for the row-mapped
        dimension.
    alpha : float, optional
        Global alpha value to use for all lines.
    join_across_missing : bool, optional
        If True, join lines across missing (NaN) data. Default is False.
    err_band_alpha : float, optional
        Alpha value to use for error bands.
    err_bar_capsize : float, optional
        Size of the caps on error bars.
    xlabel : str, optional
        Alternate label to use for the x-axis.
    ylabel : str, optional
        Alternate label to use for the y-axis.
    xlim : tuple, optional
        Limits to use for the x-axis.
    ylim : tuple, optional
        Limits to use for the y-axis.
    xscale : str, optional
        Scale to use for the x-axis, e.g. 'log'.
    yscale : str, optional
        Scale to use for the y-axis, e.g. 'log'.
    zscale : str, optional
        Scale to use for a heatmap color dimension, e.g. 'log'.
    xbase : float, optional
        If ``xscale=='log'``, the log base to use for the x-axis.
    ybase : float, optional
        If ``yscale=='log'``, the log base to use for the y-axis.
    xticks : sequence[float], optional
        Manual sequence of x-values to use for ticks.
    yticks : sequence[float], optional
        Manual sequence of y-values to use for ticks.
    xticklabels : sequence[str], optional
        Manual sequence of x-tick labels to use, requires and should be the
        same length as ``xticks``.
    yticklabels : sequence[str], optional
        Manual sequence of y-tick labels to use, requires and should be the
        same length as ``yticks``.
    vspans : sequence[float], optional
        Sequence of x-values to use for vertical spans.
    hspans : sequence[float], optional
        Sequence of y-values to use for horizontal spans.
    span_color : str or tuple, optional
        Color to use for spans.
    span_alpha : float, optional
        Alpha value to use for spans.
    span_linewidth : float, optional
        Line width to use for spans.
    span_linestyle : str, optional
        Line style to use for spans.
    grid : bool, optional
        Whether to show grid lines.
    grid_which : str, optional
        Which grid lines to show, either 'major' or 'minor'.
    grid_alpha : float, optional
        Alpha value to use for grid lines.
    legend : bool, optional
        Whether to show a legend.
    legend_ncol : int, optional
        Number of columns to use for the legend.
    legend_merge : bool, optional
        If ``True``, combinations of different mapped properties are merged
        into list of every combination.
    legend_reverse : bool, optional
        If ``True``, reverse the order of the legend entries.
    legend_entries : sequence, optional
        An explicit sequence of legend entries to use.
    legend_labels : sequence, optional
        An explicit sequence of legend labels to use.
    legend_extras : sequence, optional
        An explicit sequence of extra legend items to add.
    legend_opts : dict, optional
        Additional keyword arguments to pass to the legend plotting function.
    title : str, optional
        A title to use for the plot.
    axs : sequence[sequence[matplotlib.Axes]], optional
        An explicit array of axes to use for the plot, it should have at least
        as many rows and columns as there are mapped dimensions.
    ax : matplotlib.Axes, optional
        Shortcut for supplying a single axes to use for the plot, can only
        supply if there is a single row and column.
    format_axs : bool, optional
        Whether to format the axes to use the neutral xyzpy style.
    figsize : tuple, optional
        Size of the figure to use if creating one (ax is axs is None). If not
        specified it is automatically computed based on the number of rows and
        columns.
    height : float, optional
        Height of each subplot. Default is 3.
    width : float, optional
        Width of each subplot. If not specified, it is automatically set to
        match ``height``. Default is None.
    hspace : float, optional
        Spacing between subplots vertically. Default is 0.12.
    wspace : float, optional
        Spacing between subplots horizontally. Default is 0.12.
    sharex : bool, optional
        Whether to share the x-axis between subplots. Default is True.
    sharey : bool, optional
        Whether to share the y-axis between subplots. Default is True.
    kwargs : dict, optional
        Additional keyword arguments to pass to the main plotting function.

    Returns
    -------
    fig : matplotlib.Figure
        Figure containing the plot (None if ``ax`` or ``axs`` is specified).
    axs : sequence[sequence[matplotlib.Axes]]
        Array of axes containing the plot.
    """
    p = Infiniplotter(ds, x, y, z, **kwargs)
    if z is not None:
        p.plot_heatmap()
    else:
        p.plot_lines()
    return p.fig, p.axs



import matplotlib
import numpy as np
import xarray as xr
from numpy.testing import assert_allclose
from pytest import fixture, mark

import xyzpy as xyz
from xyzpy.plot.color import convert_colors

matplotlib.use("Template")


@fixture
def dataset_3d():
    x = [1, 2, 3, 4, 5, 6, 8]
    z = [10, 20, 40, 80]
    d = np.random.rand(7, 4)
    ds = xr.Dataset()
    ds["x"] = x
    ds["z"] = z
    ds["y"] = (("x", "z"), d)
    return ds


@fixture
def dataset_heatmap():
    x = np.linspace(10, 20, 11)
    y = np.linspace(20, 40, 21)
    xx, yy = np.meshgrid(x, y)
    c = np.cos(((xx**2 + yy**2) ** 0.5) / 2)
    s = np.sin(((xx**2 + yy**2) ** 0.5) / 2)
    ds = xr.Dataset(
        coords={"x": x, "y": y},
        data_vars={"c": (("y", "x"), c), "s": (("y", "x"), s)},
    )
    return ds.where(ds.y < 35)


@fixture
def dataset_4d():
    x = np.linspace(10, 20, 11)
    y = np.linspace(20, 40, 21)
    phi = np.linspace(-0.1, 0.1, 3)
    xx, yy, phis = np.meshgrid(x, y, phi)
    c = np.cos(((xx**2 + yy**2) ** 0.5) / 2 + phis)
    s = np.sin(((xx**2 + yy**2) ** 0.5) / 2 + phis)
    ds = xr.Dataset(
        coords={"x": x, "y": y, "phi": phi},
        data_vars={"c": (("y", "x", "phi"), c), "s": (("y", "x", "phi"), s)},
    )
    return ds


@fixture
def dataset_5d():
    x = np.linspace(10, 20, 11)
    y = np.linspace(20, 40, 21)
    phi = np.linspace(-0.5, 0.5, 3)
    A = [1, 2]
    xx, yy, phis, AA = np.meshgrid(x, y, phi, A)
    c = AA * np.cos(((xx**2 + yy**2) ** 0.5) / 2 + phis)
    s = AA * np.sin(((xx**2 + yy**2) ** 0.5) / 2 + phis)
    ds = xr.Dataset(
        coords={"x": x, "y": y, "phi": phi, "A": A},
        data_vars={
            "c": (("y", "x", "phi", "A"), c),
            "s": (("y", "x", "phi", "A"), s),
        },
    )
    return ds


@fixture
def dataset_scatter():
    x = np.random.randn(100, 100)
    y = np.random.randn(100, 100)
    z = np.random.randn(100)

    a = np.linspace(0, 3, 100)
    b = np.linspace(0, 3, 100)

    ds = xr.Dataset(
        coords={"a": a, "b": b},
        data_vars={"x": (("a", "b"), x), "y": (("a", "b"), y), "z": ("b", z)},
    )
    return ds.where(ds.z > 0.5)


# --------------------------------------------------------------------------- #
# TEST COLORS                                                                 #
# --------------------------------------------------------------------------- #


class TestConvertColors:
    def test_simple(self):
        cols = [(1, 0, 0, 1), (0, 0.5, 0, 0.5)]
        new_cols = list(convert_colors(cols, outformat="BOKEH"))
        assert new_cols == [(255, 0, 0, 1), (0, 127, 0, 0.5)]


# --------------------------------------------------------------------------- #
# TEST PLOTTERS                                                               #
# --------------------------------------------------------------------------- #


@mark.parametrize("plot_fn", [xyz.lineplot, xyz.ilineplot])
class TestCommonInterface:
    @mark.parametrize("colors", [True, False, None])
    @mark.parametrize("markers", [True, False, None])
    def test_works_2d(self, plot_fn, dataset_3d, colors, markers):
        plot_fn(
            dataset_3d,
            "x",
            "y",
            "z",
            return_fig=True,
            colors=colors,
            markers=markers,
        )

    @mark.parametrize("colormap", ["xyz", "viridis"])
    @mark.parametrize("colormap_log", [True, False])
    @mark.parametrize("colormap_reverse", [True, False])
    @mark.parametrize("string_z_coo", [True, False])
    def test_color_options(
        self,
        plot_fn,
        dataset_3d,
        string_z_coo,
        colormap,
        colormap_log,
        colormap_reverse,
    ):
        if string_z_coo:
            dataset_3d = dataset_3d.copy(deep=True)
            dataset_3d["z"] = ["a", "b", "c", "d"]
        plot_fn(
            dataset_3d,
            "x",
            "y",
            "z",
            return_fig=True,
            colors=True,
            colormap=colormap,
            colormap_log=colormap_log,
            colormap_reverse=colormap_reverse,
        )

    @mark.parametrize("markers", [True, False, None])
    def test_works_1d(self, plot_fn, dataset_3d, markers):
        plot_fn(
            dataset_3d.loc[{"z": 40}],
            "x",
            "y",
            return_fig=True,
            markers=markers,
        )

    @mark.parametrize("padding", [None, 0.1])
    @mark.parametrize("xlims", [None, (0.0, 10.0)])
    @mark.parametrize("ylims", [None, (-1.0, 1.0)])
    def test_plot_range(self, dataset_3d, plot_fn, padding, xlims, ylims):
        plot_fn(
            dataset_3d,
            "x",
            "y",
            "z",
            return_fig=True,
            padding=padding,
            xlims=xlims,
            ylims=ylims,
        )

    @mark.parametrize("xlog", [True, False])
    @mark.parametrize("ylog", [True, False])
    @mark.parametrize(
        "xticks",
        [
            None,
            (
                2,
                3,
            ),
        ],
    )
    @mark.parametrize(
        "yticks",
        [
            None,
            (
                0.2,
                0.3,
            ),
        ],
    )
    @mark.parametrize(
        "vlines",
        [
            None,
            (
                2,
                3,
            ),
        ],
    )
    @mark.parametrize(
        "hlines",
        [
            None,
            (
                0.2,
                0.3,
            ),
        ],
    )
    def test_ticks_and_lines(
        self, dataset_3d, plot_fn, xlog, ylog, xticks, yticks, vlines, hlines
    ):
        plot_fn(
            dataset_3d,
            "x",
            "y",
            "z",
            return_fig=True,
            xticks=xticks,
            yticks=yticks,
            vlines=vlines,
            hlines=hlines,
        )


class TestLinePlot:
    def test_auto_lineplot(self):
        x = [1, 2, 3]
        y = [[4, 5, 6], [7, 8, 9]]
        xyz.auto_lineplot(x, y)

    def test_multi_plot_4d(self, dataset_4d):
        dataset_4d.xyz.lineplot("x", "c", "y", row="phi")
        dataset_4d.xyz.lineplot("x", "c", "y", col="phi")

    @mark.parametrize("colors", [False, True])
    def test_multi_plot_5d(self, dataset_5d, colors):
        kws = {"colors": colors}
        dataset_5d.xyz.lineplot("x", "c", "y", row="phi", col="A", **kws)
        dataset_5d.xyz.lineplot("x", "c", "y", col="phi", row="A", **kws)


class TestILinePlot:
    def test_auto_ilineplot(self):
        x = [1, 2, 3]
        y = [[4, 5, 6], [7, 8, 9]]
        xyz.auto_ilineplot(x, y)

    def test_multi_plot_4d(self, dataset_4d):
        dataset_4d.xyz.ilineplot("x", "c", "y", row="phi")
        dataset_4d.xyz.ilineplot("x", "c", "y", col="phi")

    @mark.parametrize("colors", [False, True])
    def test_multi_plot_5d(self, dataset_5d, colors):
        kws = {"colors": colors}
        dataset_5d.xyz.ilineplot("x", "c", "y", row="phi", col="A", **kws)
        dataset_5d.xyz.ilineplot("x", "c", "y", col="phi", row="A", **kws)


class TestHeatmap:
    def test_simple(self, dataset_heatmap):
        dataset_heatmap.xyz.plot(x="x", y="y", z="c")

    def test_vspans_dict_label_uses_data_x_axes_y(self, dataset_heatmap):
        import matplotlib.transforms as transforms

        fig, axs = dataset_heatmap.xyz.plot(
            x="x",
            y="y",
            z="c",
            legend=False,
            vspans={"origin": 15.0},
        )

        ax = axs[0, 0]
        (text,) = ax.texts

        expected = transforms.blended_transform_factory(
            ax.transData,
            ax.transAxes,
        ).transform((15.0, 0.0))
        actual = text.get_transform().transform(text.get_position())

        assert text.get_text() == "origin"
        assert_allclose(actual, expected)

    def test_multi_plot_4d(self, dataset_4d):
        dataset_4d.xyz.plot("x", "y", "c", row="phi")
        dataset_4d.xyz.plot("x", "y", "c", col="phi")

    def test_multi_plot_5d(self, dataset_5d):
        dataset_5d.xyz.plot("x", "y", "c", row="phi", col="A")
        dataset_5d.xyz.plot("x", "y", "c", col="phi", row="A")


class TestIHeatmap:
    def test_simple(self, dataset_heatmap):
        dataset_heatmap.xyz.iheatmap("x", "y", "c", return_fig=True)

    def test_multi_plot_4d(self, dataset_4d):
        dataset_4d.xyz.iheatmap("x", "y", "c", row="phi")
        dataset_4d.xyz.iheatmap("x", "y", "c", col="phi")

    def test_multi_plot_5d(self, dataset_5d):
        dataset_5d.xyz.iheatmap("x", "y", "c", row="phi", col="A")
        dataset_5d.xyz.iheatmap("x", "y", "c", col="phi", row="A")


class TestScatter:
    def test_normal(self, dataset_scatter):
        dataset_scatter.xyz.scatter("x", "y")

    def test_multi_plot_4d(self, dataset_4d):
        dataset_4d.xyz.scatter("x", "y", c="c", row="phi")
        dataset_4d.xyz.scatter("x", "y", c="c", col="phi")

    def test_multi_plot_5d(self, dataset_5d):
        dataset_5d.xyz.scatter("x", "y", c="c", row="phi", col="A")
        dataset_5d.xyz.scatter("x", "y", c="c", col="phi", row="A")


class TestIScatter:
    def test_normal(self, dataset_scatter):
        dataset_scatter.xyz.iscatter("x", "y")

    def test_multi_plot_4d(self, dataset_4d):
        dataset_4d.xyz.iscatter("x", "y", c="c", row="phi")
        dataset_4d.xyz.iscatter("x", "y", c="c", col="phi")

    def test_multi_plot_5d(self, dataset_5d):
        dataset_5d.xyz.iscatter("x", "y", c="c", row="phi", col="A")
        dataset_5d.xyz.iscatter("x", "y", c="c", col="phi", row="A")


class TestHistogram:
    def test_normal(self, dataset_3d):
        dataset_3d.xyz.histogram("y", z="z")

    def test_multi_hist(self, dataset_heatmap):
        dataset_heatmap.xyz.histogram(("c", "s"))

    def test_multi_plot_4d(self, dataset_4d):
        dataset_4d.xyz.histogram(("c", "s"), row="phi")
        dataset_4d.xyz.histogram(("c", "s"), col="phi")

    def test_multi_plot_5d(self, dataset_5d):
        dataset_5d.xyz.histogram(("c", "s"), row="phi", col="A")
        dataset_5d.xyz.histogram(("c", "s"), col="phi", row="A")


class TestVisualizeMatrix:
    def test_single_real(self):
        x = np.random.randn(10, 10)
        xyz.visualize_matrix(x)

    def test_single_complex(self):
        x = np.random.randn(10, 10) + 1j * np.random.randn(10, 10)
        xyz.visualize_matrix(x)


class TestVisualizeTensor:
    def test_basic(self):
        x = np.random.randn(2, 3, 4, 5)
        xyz.visualize_tensor(x, compass=True, legend=True)

    @mark.parametrize("max_projections", [None, 1, 2, 3])
    def test_max_projections(self, max_projections):
        x = np.random.randn(2, 3, 4, 5)
        xyz.visualize_tensor(
            x, max_projections=max_projections, compass=True, legend=True
        )



"""
Helper functions for generating color spectrums.
"""

import itertools
import math

import matplotlib as mpl
import matplotlib.cm as cm
import numpy as np
from matplotlib.colors import (
    BASE_COLORS,
    CSS4_COLORS,
    Colormap,
    LinearSegmentedColormap,
    hsv_to_rgb,
    to_rgba,
)

from .xyz_cmaps import _XYZ_CMAPS


def _COLORS_MATPLOTLIB_TO_BOKEH(cols):
    for col in cols:
        yield (int(255 * col[0]), int(255 * col[1]), int(255 * col[2]), col[3])


_COLOR_CONVERT_METHODS = {
    ("MATPLOTLIB", "BOKEH"): _COLORS_MATPLOTLIB_TO_BOKEH,
}


def convert_colors_string_to_tuple(cols):
    for col in cols:
        if isinstance(col, str):
            if col in BASE_COLORS:
                yield to_rgba(BASE_COLORS[col])
            elif col in CSS4_COLORS:
                yield to_rgba(CSS4_COLORS[col])
            else:
                raise ValueError(
                    "Color specifier {} not recognized".format(col)
                )
        else:
            yield to_rgba(col)


def convert_colors(cols, outformat, informat="MATPLOTLIB"):
    """Convert lists of colors between formats"""
    # make sure cols are numeric tuples first
    cols = convert_colors_string_to_tuple(cols)

    if informat == outformat:
        return cols
    return _COLOR_CONVERT_METHODS[(informat, outformat)](cols)


def xyz_colormaps(name=None, reverse=False):
    """Custom-defined colormaps"""
    if isinstance(name, Colormap):
        if reverse:
            return name.reversed()
        else:
            return name

    try:
        import colorcet

        found_colorcet = True
    except (ImportError, KeyError):
        found_colorcet = False
    try:
        import cmocean
    except (ImportError, KeyError):
        pass

    # the default colormap
    if name is None:
        if found_colorcet:
            name = "rainbow"
        else:
            name = "xyz"

    # specify some aliases
    name = {
        "rainbow_pink": "rainbow_bgyrm_35_85_c69",
        "rainbow_pink_r": "rainbow_bgyrm_35_85_c69_r",
        "isolum_pink": "isoluminant_cm_70_c39",
        "isolum_pink_r": "isoluminant_cm_70_c39_r",
    }.get(name, name)

    # TODO: make this more general - not reliant on '_r' versions of cmaps
    if reverse:
        if name[-2:] != "_r":
            name = name + "_r"
        # 'un'-reverse
        else:
            name = name[:-2]

    # Custom xyzpy colormaps
    if name in _XYZ_CMAPS:
        return _XYZ_CMAPS[name]

    # special cases with name conflicts -->  prefer matplotlib
    if name not in {"inferno", "coolwarm", "blues"}:
        try:
            return colorcet.cm[name]
        except (NameError, KeyError):
            pass
        try:
            return getattr(cmocean.cm, name)
        except (NameError, AttributeError):
            pass

    # matplotlib colormaps
    return getattr(cm, name)


def get_default_sequential_cm(engine="MATPLOTLIB"):
    if engine == "BOKEH":
        from bokeh.palettes import Category10_9

        return itertools.cycle(Category10_9)
    else:
        return itertools.cycle(rgb + (1.0,) for rgb in cm.tab10.colors)


def cimple(
    hue,
    sat1=0.4,
    sat2=1.0,
    val1=0.95,
    val2=0.35,
    hue_shift=0.0,
    name="cimple",
    auto_adjust_sat=0.2,
):
    """Creates a color map for a single hue."""
    # account for the fact that yellows appear much less saturated
    sat1 += auto_adjust_sat * math.cos(hue * math.pi - 0.15) ** 8

    c1 = hsv_to_rgb((hue % 1.0, sat1, val1))
    c2 = hsv_to_rgb(((hue + hue_shift) % 1.0, sat2, val2))
    cdict = {
        "red": [(0.0, c1[0], c1[0]), (1.0, c2[0], c2[0])],
        "green": [(0.0, c1[1], c1[1]), (1.0, c2[1], c2[1])],
        "blue": [(0.0, c1[2], c1[2]), (1.0, c2[2], c2[2])],
    }
    return LinearSegmentedColormap(name, cdict)


def cimple_bright(
    hue,
    sat1=0.8,
    sat2=0.9,
    val1=0.97,
    val2=0.3,
    hue_shift=0.0,
    name="cimple_bright",
):
    """Creates a color map for a single hue, with bright defaults."""
    return cimple(hue, sat1, sat2, val1, val2, hue_shift, name)


def cimluv(
    hue,
    hue_shift=0.0,
    sat1=1.00,
    sat2=0.50,
    val1=0.80,
    val2=0.30,
    N=30,
    reverse=False,
):
    """Creates a color map for single hue, using HSLuv color space."""
    import hsluv
    import numpy as np

    pos = np.linspace(0.0, 1.0, N)
    hue1 = hue - hue_shift / 2
    hue2 = hue + hue_shift / 2
    hues = np.linspace(360 * hue1, 360 * hue2, N)
    sats = np.linspace(100 * sat1, 100 * sat2, N)
    luvs = np.linspace(100 * val1, 100 * val2, N)
    cdict = {"red": [], "green": [], "blue": []}
    for p, h, s, l in zip(pos, hues, sats, luvs):
        r, g, b = hsluv.hsluv_to_rgb((h, s, l))
        cdict["red"].append((p, r, r))
        cdict["green"].append((p, g, g))
        cdict["blue"].append((p, b, b))

    if reverse:
        for key in cdict:
            cdict[key].reverse()

    return LinearSegmentedColormap("cimluv", cdict)


def oklch_to_oklab(l, c, h):
    h_rad = np.radians(h)
    a = c * np.cos(h_rad)
    b = c * np.sin(h_rad)
    return (l, a, b)


def oklab_to_xyz(l, a, b):
    l_prime = l + 0.3963377774 * a + 0.2158037573 * b
    m_prime = l - 0.1055613458 * a - 0.0638541728 * b
    s_prime = l - 0.0894841775 * a - 1.2914855480 * b

    l_linear = l_prime**3
    m_linear = m_prime**3
    s_linear = s_prime**3

    x = (
        1.227013851103521 * l_linear
        - 0.5577999806518222 * m_linear
        + 0.2812561489664678 * s_linear
    )
    y = (
        -0.04058017842328059 * l_linear
        + 1.1122568696168302 * m_linear
        - 0.0716766786656012 * s_linear
    )
    z = (
        -0.0763812845057069 * l_linear
        - 0.4214819784180127 * m_linear
        + 1.586163220440795 * s_linear
    )
    return (x, y, z)


def xyz_to_linear_rgb(x, y, z):
    r = 3.2404542 * x - 1.5371385 * y - 0.4985314 * z
    g = -0.9692660 * x + 1.8760108 * y + 0.0415560 * z
    b = 0.0556434 * x - 0.2040259 * y + 1.0572252 * z

    r = np.clip(r, 0.0, 1.0)
    g = np.clip(g, 0.0, 1.0)
    b = np.clip(b, 0.0, 1.0)
    return (r, g, b)


def linear_rgb_to_srgb(r, g, b):
    def gamma(u):
        return np.where(
            u <= 0.0031308,
            12.92 * u,
            1.055 * (u ** (1 / 2.4)) - 0.055,
        )

    sr = gamma(r)
    sg = gamma(g)
    sb = gamma(b)
    return (sr, sg, sb)


def oklch_to_rgb(l, c, h):
    l_ok, a_ok, b_ok = oklch_to_oklab(l, c, h)
    x, y, z = oklab_to_xyz(l_ok, a_ok, b_ok)
    r_linear, g_linear, b_linear = xyz_to_linear_rgb(x, y, z)
    sr, sg, sb = linear_rgb_to_srgb(r_linear, g_linear, b_linear)
    sr = np.clip(sr, 0.0, 1.0)
    sg = np.clip(sg, 0.0, 1.0)
    sb = np.clip(sb, 0.0, 1.0)
    return (sr, sg, sb)


def srgb_to_linear_rgb(sr, sg, sb):
    def inverse_gamma(u):
        if u <= 0.04045:
            return u / 12.92
        else:
            return ((u + 0.055) / 1.055) ** 2.4

    r_linear = inverse_gamma(sr)
    g_linear = inverse_gamma(sg)
    b_linear = inverse_gamma(sb)
    return (r_linear, g_linear, b_linear)


def linear_rgb_to_xyz(r, g, b):
    x = 0.4124564 * r + 0.3575761 * g + 0.1804375 * b
    y = 0.2126729 * r + 0.7151522 * g + 0.0721750 * b
    z = 0.0193339 * r + 0.1191920 * g + 0.9503041 * b
    return (x, y, z)


def xyz_to_oklab(x, y, z):
    l = 0.8189330101 * x + 0.3618667424 * y - 0.1288597137 * z
    m = 0.0329845436 * x + 0.9293118715 * y + 0.0361456387 * z
    s = 0.0482003018 * x + 0.2643662691 * y + 0.6338517070 * z

    l_prime = l ** (1 / 3)
    m_prime = m ** (1 / 3)
    s_prime = s ** (1 / 3)

    L = (
        0.2104542553 * l_prime
        + 0.7936177850 * m_prime
        - 0.0040720468 * s_prime
    )
    a = (
        1.9779984951 * l_prime
        - 2.4285922050 * m_prime
        + 0.4505937099 * s_prime
    )
    b = (
        0.0259040371 * l_prime
        + 0.7827717662 * m_prime
        - 0.8086757660 * s_prime
    )
    return (L, a, b)


def oklab_to_oklch(L, a, b):
    c = math.sqrt(a**2 + b**2)
    h_rad = math.atan2(b, a)
    h_deg = math.degrees(h_rad) % 360.0
    return (L, c, h_deg)


def rgb_to_oklch(sr, sg, sb):
    r_linear, g_linear, b_linear = srgb_to_linear_rgb(sr, sg, sb)
    x, y, z = linear_rgb_to_xyz(r_linear, g_linear, b_linear)
    L, a, b = xyz_to_oklab(x, y, z)
    l, c, h = oklab_to_oklch(L, a, b)
    return (l, c, h)


def cmoke(
    hue,
    hue_shift=0.0,
    sat1=0.36,
    sat2=0.50,
    val1=0.38,
    val2=0.93,
    N=51,
    reverse=False,
):
    """Creates a color map for single hue, using OKLCH color space."""
    import numpy as np

    if isinstance(hue, (list, tuple)):
        hue1, hue2 = hue
    else:
        hue1 = hue - hue_shift / 2
        hue2 = hue + hue_shift / 2

    hs = np.linspace(360 * hue1, 360 * hue2, N)
    ls = np.linspace(val1, val2, N)
    cs = np.linspace(0.37 * sat1, 0.37 * sat2, N)

    pos = np.linspace(0.0, 1.0, N)
    cdict = {"red": [], "green": [], "blue": []}

    for p, l, c, h in zip(pos, ls, cs, hs):
        r, g, b = oklch_to_rgb(l, c, h)
        cdict["red"].append((p, r, r))
        cdict["green"].append((p, g, g))
        cdict["blue"].append((p, b, b))

    cmap = LinearSegmentedColormap("cmoke", cdict)

    if reverse:
        cmap = cmap.reversed()

    return cmap


doublerainbow = cmoke((1.8, -0.1), val1=0.2)
doublerainbow_r = doublerainbow.reversed()

mpl.colormaps.register(cmap=doublerainbow, name="doublerainbow")
mpl.colormaps.register(cmap=doublerainbow_r, name="doublerainbow_r")



"""Functions for systematically evaluating a function over specific cases."""

import functools
import itertools

import numpy as np
import xarray as xr

from .combo_runner import (
    combo_runner_core,
    combo_runner_to_ds,
)
from .prepare import (
    parse_cases,
    parse_combos,
    parse_constants,
    parse_fn_args,
    parse_resources,
    parse_var_coords,
    parse_var_dims,
    parse_var_names,
)


def case_runner(
    fn,
    fn_args,
    cases,
    combos=None,
    constants=None,
    split=False,
    shuffle=False,
    parse=True,
    parallel=False,
    executor=None,
    num_workers=None,
    verbosity=1,
):
    """Simple case runner that outputs the raw tuple of results.

    Parameters
    ----------
    fn : callable
        Function with which to evalute cases with
    fn_args : tuple
        Names of case arguments that fn takes, can be ``None`` if each case is
        a ``dict``.
    cases : iterable[tuple] or iterable[dict]
        List of specific configurations that ``fn_args`` should take. If
        ``fn_args`` is ``None``, each case should be a ``dict``.
    combos : dict_like[str, iterable], optional
        Optional specification of sub-combinations.
    constants : dict, optional
        Constant function arguments.
    split : bool, optional
        See :func:`~xyzpy.combo_runner`.
    shuffle : bool or int, optional
        If given, compute the results in a random order (using ``random.seed``
        and ``random.shuffle``), which can be helpful for distributing
        resources when not all cases are computationally equal.
    parallel : bool, optional
        Process combos in parallel, default number of workers picked.
    executor : executor-like pool, optional
        Submit all combos to this pool executor. Must have ``submit`` or
        ``apply_async`` methods and API matching either ``concurrent.futures``
        or an ``ipyparallel`` view. Pools from ``multiprocessing.pool`` are
        also  supported.
    num_workers : int, optional
        Explicitly choose how many workers to use, None for automatic.
    verbosity : {0, 1, 2}, optional
        How much information to display:

            - 0: nothing,
            - 1: just progress,
            - 2: all information.

    Returns
    -------
        results : list of fn output for each case
    """
    if parse:
        # Prepare fn_args and values
        fn_args = parse_fn_args(fn, fn_args)
        cases = parse_cases(cases, fn_args)
        combos = parse_combos(combos)
        constants = parse_constants(constants)

    return combo_runner_core(
        fn,
        cases=cases,
        combos=combos,
        constants=constants,
        parallel=parallel,
        num_workers=num_workers,
        executor=executor,
        verbosity=verbosity,
        split=split,
        flat=True,
        shuffle=shuffle,
    )


def case_runner_to_ds(
    fn,
    fn_args,
    cases,
    var_names,
    var_dims=None,
    var_coords=None,
    combos=None,
    constants=None,
    resources=None,
    attrs=None,
    shuffle=False,
    to_df=False,
    parse=True,
    parallel=False,
    num_workers=None,
    executor=None,
    verbosity=1,
):
    """Takes a list of ``cases`` to run ``fn`` over, possibly in parallel, and
    outputs a :class:`xarray.Dataset`.

    Parameters
    ----------
    fn : callable
        Function to evaluate.
    fn_args : str or iterable[str]
        Names and order of arguments to ``fn``, can be ``None`` if ``cases``
        are supplied as dicts.
    cases: iterable[tuple] or iterable[dict]
        List of configurations used to generate results.
    var_names : str or iterable of str
        Variable name(s) of the output(s) of ``fn``.
    var_dims : sequence of either strings or string sequences, optional
        'Internal' names of dimensions for each variable, the values for
        each dimension should be contained as a mapping in either
        `var_coords` (not needed by `fn`) or `constants` (needed by `fn`).
    var_coords : mapping, optional
        Mapping of extra coords the output variables may depend on.
    combos : dict_like[str, iterable], optional
        If specified, run all combinations of some arguments in these mappings.
    constants : mapping, optional
        Arguments to `fn` which are not iterated over, these will be
        recorded either as attributes or coordinates if they are named
        in `var_dims`.
    resources : mapping, optional
        Like `constants` but they will not be recorded.
    attrs : mapping, optional
        Any extra attributes to store.
    shuffle : bool or int, optional
        If given, compute the results in a random order (using ``random.seed``
        and ``random.shuffle``), which can be helpful for distributing
        resources when not all cases are computationally equal.
    parse : bool, optional
        Whether to perform parsing of the inputs arguments.
    parallel : bool, optional
        Process combos in parallel, default number of workers picked.
    executor : executor-like pool, optional
        Submit all combos to this pool executor. Must have ``submit`` or
        ``apply_async`` methods and API matching either ``concurrent.futures``
        or an ``ipyparallel`` view. Pools from ``multiprocessing.pool`` are
        also  supported.
    num_workers : int, optional
        Explicitly choose how many workers to use, None for automatic.
    verbosity : {0, 1, 2}, optional
        How much information to display:

            - 0: nothing,
            - 1: just progress,
            - 2: all information.

    Returns
    -------
    ds : xarray.Dataset
        Dataset with minimal covering coordinates and all cases
        evaluated.
    """
    if parse:
        # Prepare fn_args and values
        fn_args = parse_fn_args(fn, fn_args)
        cases = parse_cases(cases, fn_args)
        combos = parse_combos(combos)
        constants = parse_constants(constants)
        resources = parse_resources(resources)
        var_names = parse_var_names(var_names)
        var_dims = parse_var_dims(var_dims, var_names=var_names)
        var_coords = parse_var_coords(var_coords)

    return combo_runner_to_ds(
        fn=fn,
        combos=combos,
        var_names=var_names,
        var_dims=var_dims,
        var_coords=var_coords,
        cases=cases,
        constants=constants,
        resources=resources,
        attrs=attrs,
        shuffle=shuffle,
        to_df=to_df,
        parallel=parallel,
        num_workers=num_workers,
        executor=executor,
        verbosity=verbosity,
        parse=False,
    )


case_runner_to_df = functools.partial(case_runner_to_ds, to_df=True)


# --------------------------------------------------------------------------- #
# Update or add new values                                                    #
# --------------------------------------------------------------------------- #


def is_case_missing(ds, setting, method="isnull"):
    """Does the dataset or dataarray ``ds`` not contain any non-null data for
    single location ``setting``?

    Note that this only returns true if *all* data across *all* variables is
    completely missing at the location.

    Parameters
    ----------
    ds : xarray.Dataset or xarray.DataArray
        Dataset or dataarray to look in.
    setting : dict[str, hashable]
        Coordinate location to check.

    Returns
    -------
    missing : bool
    """
    import numpy as np

    try:
        sds = ds.sel(setting)

        if method == "isnull":
            sds = sds.isnull()
        elif method == "isfinite":
            sds = ~np.isfinite(sds)
        else:
            raise ValueError("Unknown method: {}".format(method))

        nds = sds.all()
    except KeyError:
        # coordinates not present at all
        return True

    try:
        # cocatenate answer for each variable if Dataset
        nds = nds.to_array().all()
    except AttributeError:
        # already a DataArray
        pass

    return nds.item()


def parse_into_cases(combos=None, cases=None, ds=None, method="isnull"):
    """Convert maybe ``combos`` and maybe ``cases`` to a single list of
    ``cases`` only, also optionally filtering based on whether any data at each
    location is already present in Dataset or DataArray ``ds``.

    Note that this only checks whether *all* data across *all* variables is
    completely missing at the location. To check against a single variable only
    simply supply a DataArray instead of a Dataset, e.g. ``ds=ds["var_name"]``.

    Parameters
    ----------
    combos : dict_like[str, iterable], optional
        Parameter combinations.
    cases : iterable[dict], optional
        Parameter configurations.
    ds : xarray.Dataset or xarray.DataArray, optional
        Dataset or DataArray in which to check for existing data.
    method : {"isnull", "isfinite"}, optional
        How to determine whether data is missing when ``ds`` is supplied.
        "isnull" checks for null/nan values, while "isfinite" checks for all
        non-finite values (i.e. inf or nan).

    Returns
    -------
    new_cases : iterable[dict]
        The combined and possibly filtered list of cases.
    """
    combos = parse_combos(combos)

    if combos:
        combo_keys, combo_values = zip(*combos)
    else:
        combo_keys, combo_values = [], []

    if cases is None:
        cases = [{}]

    new_cases = []

    if ds is None:
        # we can just flatten all cases and combos
        for case in cases:
            for combo in itertools.product(*combo_values):
                setting = case | dict(zip(combo_keys, combo))
                new_cases.append(setting)
        return new_cases

    # else we need to check against existing data
    existing_coords = {
        dim: {v: i for i, v in enumerate(ds.coords[dim].values)}
        for dim in ds.dims
    }

    # first we sort into cases which are outside of the existing coordinates
    cases_inside = []
    ilocs = []
    for case in cases:
        for combo in itertools.product(*combo_values):
            setting = case | dict(zip(combo_keys, combo))
            # build its index, or break if outside
            iloc = []
            for dim, val in setting.items():
                idx = existing_coords[dim].get(val, None)
                if idx is None:
                    # missing value, don't need to check actual data
                    new_cases.append(setting)
                    break
                iloc.append(idx)
            else:
                # no break - location is in bounds
                cases_inside.append(setting)
                ilocs.append(iloc)

    if not cases_inside:
        # no cases inside, we're done
        return new_cases

    # now we check if the actual data for the cases inside is finite
    # this fancy index extracts the values at the inside case locations
    indices = tuple(np.array(col) for col in zip(*ilocs))

    if isinstance(ds, xr.DataArray):
        ds = ds.to_dataset()

    if method == "isnull":
        ds = ds.isnull()
    elif method == "isfinite":
        ds = ~np.isfinite(ds)
    else:
        raise ValueError("Unknown method: {}".format(method))

    missing = None
    for v in ds.data_vars:
        missing_v = ds[v].values[indices]
        if missing_v.ndim > 1:
            # sub-coordinates, we need to reduce over
            missing_v = missing_v.reshape(missing_v.shape[0], -1).any(axis=1)

        # reduce across variables, requiring *all* to be missing
        if missing is None:
            missing = missing_v
        else:
            missing = missing & missing_v

    # turn into indices of missing cases, and add to missing list
    new_cases.extend(cases_inside[i] for i in np.nonzero(missing)[0])

    return new_cases


def find_missing_cases(ds, ignore_dims=None, method="isnull"):
    """Find all cases in a dataset or DataArray with missing data.

    Parameters
    ----------
    ds : xarray.Dataset or xarray.DataArray
        Dataset or DataArray in which to find missing data
    ignore_dims : set, optional
        Internal variable dimensions (i.e. to ignore). By default (None) this
        is set to any dimensions that don't appear on all variables.

    Returns
    -------
    cases_missing : iterable[dict]
        List of cases with missing data, where each case is a dict mapping from
        dimension name to coordinate value.
    """
    if isinstance(ds, xr.DataArray):
        ds = ds.to_dataset()

    if ignore_dims is None:
        # default to ignoring any dimensions that don't appear on all variables
        ignore_dims = set(
            dim
            for dim in ds.dims
            if not all(dim in ds[v].dims for v in ds.data_vars)
        )
    elif isinstance(ignore_dims, str):
        ignore_dims = {ignore_dims}
    elif ignore_dims:
        ignore_dims = set(ignore_dims)
    else:
        ignore_dims = set()

    combos = {
        dim: ds.coords[dim].values for dim in ds.dims if dim not in ignore_dims
    }

    return parse_into_cases(combos=combos, ds=ds, method=method)
